#! /usr/local/bin/ruby
# encoding: utf-8
#
# pgx.rb
# PostgreSQL (libpq) の操作。
#

require "fiddle/import";
require "bigdecimal";
require "date";

# ======================================================================
#
class PgException < RuntimeError
end

# ======================================================================
#
module PGX
	extend	Fiddle::Importer;
	dlload	"libpq.so";

	typealias "PGconn", "void";
	typealias "PGresult", "void";
	typealias "ConnStatusType", "int";
	typealias "Oid", "int";
	typealias "ExecStatusType", "int";
	typealias "size_t", "int";

	extern	"PGconn * PQconnectdb(const char *)";	# 31.1
	extern	"void PQfinish(PGconn *)";		# 31.1
	extern	"ConnStatusType PQstatus(const PGconn *)";	# 31.2
	extern	"char * PQerrorMessage(const PGconn *)";	# 31.2
	extern	"PGresult * PQexecParams(PGconn *, const char *, int, const Oid *, const char * const *, const int *, const int *, int)";	# 31.3.1
	extern	"ExecStatusType PQresultStatus(const PGresult *)";	# 31.3.1
	extern	"char * PQresultErrorMessage(const PGresult *)";	# 31.3.1
	extern	"void PQclear(PGresult *)";	# 31.3.1
	extern	"int PQntuples(const PGresult *)";	# 31.3.2
	extern	"int PQnfields(const PGresult *)";	# 31.3.2
	extern	"char * PQfname(const PGresult *, int)";	# 31.3.2
	extern	"Oid PQftype(const PGresult *, int)";	# 31.3.2
	extern	"char * PQgetvalue(const PGresult *, int, int)";	# 31.3.2
	extern	"int PQgetisnull(const PGresult *, int, int)";	# 31.3.2
	extern	"int PQgetlength(const PGresult *, int, int)";	# 31.3.2
end

# ======================================================================
#
class Pq

	# --------------------------------------------------------------
	# enum ExecStatusType (src/interfaces/libpq/libpq-fe.h)
	#
	PGRES_EMPTY_QUERY = 0;
	PGRES_COMMAND_OK = 1;		# データを返さないコマンドが正常終了
	PGRES_TUPLES_OK = 2;		# データを返すコマンドが正常終了

	# --------------------------------------------------------------
	# enum ConnStatusType (src/interfaces/libpq/libpq-fe.h)
	#
	CONNECTION_OK = 0;
	CONNECTION_BAD = 1;

	# --------------------------------------------------------------
	# 組み込みデータ型を表すOID (src/include/catalog/pg_type.h)
	#
	BOOLOID = 16;
	BYTEAOID = 17;
	INT8OID = 20;
	INT2OID = 21;
	INT4OID = 23;
	TEXTOID = 25;
	FLOAT4OID = 700;
	FLOAT8OID = 701;
	BPCHAROID = 1042;
	VARCHAROID = 1043;
	DATEOID = 1082;
	TIMEOID = 1083;
	TIMESTAMPOID = 1114;
	TIMESTAMPTZOID = 1184;
	INTERVALOID = 1186;
	BITOID = 1560;
	VARBITOID = 1562;
	NUMERICOID = 1700;

	# ==============================================================
	# conninfo: 'postgresql://localhost/db_plrb' など。
	#
	def initialize(conninfo)
		@pq = PGX::PQconnectdb(conninfo);
		if PGX::PQstatus(@pq) != CONNECTION_OK
			PGX::PQfinish(@pq);
			@pq = nil;
			raise PgException, "Pq: #{conninfo.inspect}: connection failed.";
		end
	end

	# ==============================================================
	#
	def disconnect()
		PGX::PQfinish(@pq);
		@pq = nil;
	end

	# ==============================================================
	#
	def last_error_message()
		return PGX::PQerrorMessage(@pq).to_s;
	end

	# ==============================================================
	# (一般には)「SELECT」以外の文を実行する。
	#
	def exec(sql, *params)
		num_params = params.size;
		pack_pattern = params.map do |param|
			(param != nil) ? 'p' : 'x4';
		end.join;
		packed_params = params.compact.map do |param|
			if param.kind_of?(BigDecimal)
				param.to_s("F").sub(/\.0$/, '').dup;
			else
				param.to_s.dup;
			end
		end.pack(pack_pattern);

		pq_res = PGX::PQexecParams(@pq, sql, num_params,
				nil, packed_params, nil, nil, 0);
		case PGX::PQresultStatus(pq_res)
		when PGRES_COMMAND_OK
			PGX::PQclear(pq_res);
			return true;
		when PGRES_TUPLES_OK
			results = pq_result_to_hash(pq_res);
			PGX::PQclear(pq_res);
			return results;
		else
			err = PGX::PQresultErrorMessage(pq_res).to_s;
			PGX::PQclear(pq_res);
			raise PgException, err;
		end
	end

	alias :execute :exec;

	# --------------------------------------------------------------
	#
	def pq_result_to_hash(pq_res)
		nfileds = PGX::PQnfields(pq_res);
		fname_array = Array.new(nfileds) do |i|
			PGX::PQfname(pq_res, i).to_s;
		end
		ftype_array = Array.new(nfileds) do |i|
			PGX::PQftype(pq_res, i);
		end

		results = Array.new(PGX::PQntuples(pq_res)) do |n|
			value = Hash.new;
			nfileds.times do |i|
				if PGX::PQgetisnull(pq_res, n, i) == 1
					value[fname_array[i]] = nil;
				else
					size = PGX::PQgetlength(pq_res, n, i);
					value[fname_array[i]] = convert_type(
						PGX::PQgetvalue(pq_res, n, i).to_s(size),
						ftype_array[i]);
				end
			end
			value;
		end
		return results;
	end

	# --------------------------------------------------------------
	#
	def convert_type(val, ftype)
		case ftype
		when BOOLOID
			value = (val == 't');
		when INT8OID, INT2OID, INT4OID
			value = val.to_i;
		when NUMERICOID
			value = BigDecimal.new(val);
		when FLOAT4OID, FLOAT8OID
			value = val.to_f;
		when TEXTOID, BPCHAROID, VARCHAROID
			value = val.to_s.force_encoding("UTF-8");
		when BYTEAOID
			value = val;
		when DATEOID
			value = Date.parse(val);
		when TIMEOID
			h, m, s = val.split(/:/).map do |nn|
				nn.to_i;
			end;
			value = Time.at(h * 3600 + m * 60 + s);
		when TIMESTAMPOID, TIMESTAMPTZOID
			value = DateTime.parse(val);
		when BITOID, VARBITOID
			value = val;
		when INTERVALOID
			value = val;
		else
			value = val;
		end
		return value;
	end

	# ==============================================================
	# 「SELECT」を実行し、1レコードを表すHashの配列を返す。
	# ## 実際には exec() と同じ。
	#
	alias :select :exec;
	alias :select_all :exec;

	# ==============================================================
	# 「SELECT」を実行し、Hashの形で1レコードのみを返す。
	# 該当するレコードがない場合は空のHash ({})。
	#
	def select_one(sql, *params)
		return select_all(sql, *params)[0] || {};
	end
end

# ======================================================================
# Pgの代替実装版
#
module Pg
	@pq = nil;

	# ==============================================================
	#
	TypeBOOL = Pq::BOOLOID;
	TypeBOOLEAN = Pq::BOOLOID;
	TypeINT8 = Pq::INT8OID;
	TypeBIGINT = Pq::INT8OID;
	TypeINT2 = Pq::INT2OID;
	TypeSMALLINT = Pq::INT2OID;
	TypeINT4 = Pq::INT4OID;
	TypeINTEGER = Pq::INT4OID;
	TypeNUMERIC = Pq::NUMERICOID;
	TypeDECIMAL = Pq::NUMERICOID;
	TypeFLOAT4 = Pq::FLOAT4OID;
	TypeREAL = Pq::FLOAT4OID;
	TypeFLOAT8 = Pq::FLOAT8OID;
	TypeDOUBLE = Pq::FLOAT8OID;
	TypeDOUBLE_PRECISION = Pq::FLOAT8OID;
	TypeTEXT = Pq::TEXTOID;
	TypeBPCHAR = Pq::BPCHAROID;
	TypeCHARACTER = Pq::BPCHAROID;
	TypeVARCHAR = Pq::VARCHAROID;
	TypeCHARACTER_VARYING = Pq::VARCHAROID;
	TypeBYTEA = Pq::BYTEAOID;
	TypeDATE = Pq::DATEOID;
	TypeTIME = Pq::TIMEOID;
	TypeTIMESTAMP = Pq::TIMESTAMPOID;
	TypeTIMESTAMPTZ = Pq::TIMESTAMPTZOID;
	TypeBIT = Pq::BITOID;
	TypeVARBIT = Pq::VARBITOID;
	TypeBIT_VARYING = Pq::VARBITOID;
	TypeINTERVAL = Pq::INTERVALOID;

	# ==============================================================
	#
	def self::connect(conninfo)
		@pq = Pq.new(conninfo);
	end

	# ==============================================================
	#
	def self::disconnect()
		@pq.disconnect();
		@pq = nil;
	end

	# ==============================================================
	#
	def self::execute(sql, *params)
		return @pq.execute(sql, *params);
	end

	# ==============================================================
	#
	def self::select(sql, *params)
		return @pq.select(sql, *params);
	end

	# ==============================================================
	#
	LevelDEBUG5 = 10;
	LevelDEBUG4 = 11;
	LevelDEBUG3 = 12;
	LevelDEBUG2 = 13;
	LevelDEBUG1 = 14;
	LevelLOG = 15;
	LevelCOMMERROR = 16;
	LevelINFO = 17;
	LevelNOTICE = 18;
	LevelWARNING = 19;
	LevelERROR = 20;
	LevelFATAL = 21;
	LevelPANIC = 22;

	LevelString = {
		LevelDEBUG5 => "DEBUG5",
		LevelDEBUG4 => "DEBUG4",
		LevelDEBUG3 => "DEBUG3",
		LevelDEBUG2 => "DEBUG2",
		LevelDEBUG1 => "DEBUG1",
		LevelLOG => "LOG",
		LevelCOMMERROR => "COMMERROR",
		LevelINFO => "INFO",
		LevelNOTICE => "NOTICE",
		LevelWARNING => "WARNING",
		LevelERROR => "ERROR",
		LevelFATAL => "FATAL",
		LevelPANIC => "PANIC",
	};
	LevelString.default = "UNKNOWN";

	# ==============================================================
	#
	def self::elog(level, message)
		$stderr.print "[#{LevelString[level]}] #{message}\n";
	end

end

