/* plrb.c */
/*
** plrb.c: PL/rb: Ruby Procedural Language to create PostgreSQL function
** Copyright (C) 2015 KOYAMA Hiro <tac@amris.co.jp>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include	<stdarg.h>

#include	"postgres.h"
#include	"fmgr.h"
#include	"funcapi.h"
#include	"pgtime.h"

#include	"access/heapam.h"
#include	"access/htup_details.h"
#include	"access/xact.h"
#include	"catalog/pg_proc.h"
#include	"catalog/pg_type.h"
#include	"commands/event_trigger.h"
#include	"commands/trigger.h"
#include	"datatype/timestamp.h"
#include	"executor/spi.h"
#include	"utils/array.h"
#include	"utils/builtins.h"
#include	"utils/date.h"
#include	"utils/elog.h"
#include	"utils/guc.h"
#include	"utils/lsyscache.h"
#include	"utils/numeric.h"
#include	"utils/rel.h"
#include	"utils/syscache.h"
#include	"utils/timestamp.h"
#include	"utils/typcache.h"
#include	"utils/varbit.h"

#include	<ruby.h>

#include	"plrb.h"

#ifdef	PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/* ---------------------------------------------------------------------
*/
static
int	x_elog(int level, const char * fmt, ...)
{
	va_list		ap;
	char		buf[256];
	int		n;

	va_start(ap, fmt);
	n = vsnprintf(buf, 256, fmt, ap);

	switch (level) {
			/* postgresql.confのログ関係設定が既定値のとき */
	case DEBUG5:
	case DEBUG4:
	case DEBUG3:
	case DEBUG2:
	case DEBUG1:		/* 出力なし */
	case LOG:		/* ログ・ファイルに出力 */
	case COMMERROR:		/* ログ・ファイルに出力 */
	case INFO:		/* コンソールに出力 */
	case NOTICE:		/* コンソールに出力 */
	case WARNING:		/* コンソールとログ・ファイルに出力 */
			/* 以下は呼び出し元に制御が戻らない */
	case ERROR:		/* コンソールとログ・ファイルに出力 */
			/* 以下はPG_TRY()ブロックで捕捉できない */
	case FATAL:		/* コンソールとログ・ファイルに出力 */
			/* 以下はさらに、PANIC処理が実行される */
	case PANIC:		/* 呼び出し元に戻らない */
		ereport(level, (errmsg_internal("%s", buf)));
		break;
	default:
		break;
	}
	va_end(ap);
	return n;
}

/* ---------------------------------------------------------------------
*/
static
int	info_out(const char * fmt, ...)
{
	va_list		ap;
	char		buf[256];
	int		n = 0;

	va_start(ap, fmt);
	n = vsnprintf(buf, 256, fmt, ap);
	elog(INFO, "%s", buf);
	va_end(ap);
	return n;
}

/* ---------------------------------------------------------------------
** トップ・レベルの函数呼び出し。
** args: Array:
**	[0]: String: 函数名。
**	[1]以降: (函数に合わせた型の) 引数。
**		これをArray型の、単一の引数として渡す。
**		函数内では opt[n] の形で参照。
*/
VALUE	x_rb_funcall(VALUE args)
{
	VALUE	func_name_value = rb_ary_shift(args);
	char *	func_name = StringValuePtr(func_name_value);
	return rb_funcall(Qnil, rb_intern(func_name), 1, args);
}

/* ---------------------------------------------------------------------
** トップ・レベルの函数呼び出し。
** args: Array:
**	[0]: String: 函数名。
**	[1]以降: (函数に合わせた型の) 引数。
**		これを個別の引数として渡す。
**		函数内では定義に現れる名前で参照。
*/
VALUE	x_rb_funcallv(VALUE args)
{
	VALUE	func_name_value = rb_ary_shift(args);
	char *	func_name = StringValuePtr(func_name_value);

	int	size = NUM2INT(rb_funcall(args, rb_intern("size"), 0));
	VALUE *	args_array = ALLOCA_N(VALUE, size);
	int	i;
	for (i = 0; i < size; i++) {
		args_array[i] = rb_ary_shift(args);
	}
	return rb_funcallv(Qnil, rb_intern(func_name), size, args_array);
}

/* ---------------------------------------------------------------------
** x_rb_funcall() で例外が発生したときの後処理。
** バックトレース等をログに出力する。
** args: String: 函数名。
** excep: 例外オブジェクト。
**	err.print excep.message;
**	excep.backtrace.each do |buf|
**		err.print "... #{buf}";
**	end
*/
VALUE	x_rb_rescue(VALUE args, VALUE excep)
{
	char *	func_name = StringValuePtr(args);
	VALUE	val_message = rb_funcall(excep, rb_intern("message"), 0);
	char *	message = StringValuePtr(val_message);
	VALUE	val_backtrace = rb_funcall(excep, rb_intern("backtrace"), 0);

	VALUE	elem;
	char *	backtrace;
	char *	prev_backtrace;
	char *	p;
	int	len;

	backtrace = palloc0(1);
	strcpy(backtrace, "");
	prev_backtrace = backtrace;
	while (! NIL_P(elem = rb_ary_shift(val_backtrace))) {
		p = StringValuePtr(elem);
		len = strlen(backtrace) + strlen(p) + 5;
		prev_backtrace = backtrace;

		backtrace = palloc0(len + 1);
		sprintf(backtrace, "%s... %s\n", prev_backtrace, p);

		pfree(prev_backtrace);
	}

	ereport(ERROR,
		(errcode(ERRCODE_EXTERNAL_ROUTINE_EXCEPTION),
		 errmsg("%s: %s\n%s", func_name, message, backtrace)));
	return Qnil;
}

/* ---------------------------------------------------------------------
** 文字列をRubyプログラムとして評価する、rb_eval_string() の呼び出し。
*/
VALUE	x_rb_eval_string(VALUE args)
{
	char *	def_func_source = StringValuePtr(args);
	return rb_eval_string(def_func_source);
}

/* ---------------------------------------------------------------------
** x_rb_eval_string() で例外が発生したときの後処理。
*/
VALUE	x_rb_eval_string_rescue(VALUE args, VALUE excep)
{
	char *	func_name = StringValuePtr(args);
	VALUE	val = rb_funcall(excep, rb_intern("message"), 0);
	char *	message = StringValuePtr(val);

	ereport(ERROR,
		(errcode(ERRCODE_EXTERNAL_ROUTINE_EXCEPTION),
		 errmsg("Syntax Error: %s\n%s", func_name, message)));

	return Qnil;
}

/* ---------------------------------------------------------------------
** rb_require() の呼び出し。
*/
VALUE	x_rb_require(VALUE args)
{
	char *	require_feature = StringValuePtr(args);
	return rb_require(require_feature);
}

/* ---------------------------------------------------------------------
** rb_require() の呼び出しで例外が発生したときの処理。
*/
VALUE	x_rb_require_rescue(VALUE args, VALUE excep)
{
	char *	require_feature = StringValuePtr(args);
	info_out("Failed to auto-require %s; ignored.", require_feature);
	return Qnil;
}

/* ---------------------------------------------------------------------
** postgresql.conf に
**	plrb.string_force_encoding = 'UTF-8'
** のような記述があれば、char/varchar/text引数から変換したStringの
** エンコーディングを自動変換する。
** 既定値: UTF-8
*/
static
VALUE	string_force_encoding(VALUE str)
{
	const char *	enc;

	if ((enc = GetConfigOption("plrb.string_force_encoding", true, false)) == NULL) {
		enc = "UTF-8";
	}
	rb_funcall(str, rb_intern("force_encoding"), 1, rb_str_new_cstr(enc));
	return str;
}

/* ---------------------------------------------------------------------
*/
static
VALUE	string_force_ascii_8bit(VALUE str)
{
	rb_funcall(str, rb_intern("force_encoding"),
				1, rb_str_new_cstr("ASCII-8BIT"));
	return str;
}

/* ---------------------------------------------------------------------
** Datum (PostgreSQL) → VALUE (Ruby) の変換。
*/
static
VALUE	DatumToVALUE(Datum datum, Oid typid)
{
	VALUE		t;
	VALUE		arg[10];
	VALUE		klass;
	char *		dec;
	text *		text_val;
	BpChar *	bp_char;
	VarChar *	var_char;
	bytea *		bytea_str;

	Timestamp	timestamp;
	TimestampTz	timestamptz;
	int		tzp;
	struct pg_tm	tm;
	fsec_t		fsec;

	Oid		typOutput;
	bool		typIsVarlena;

	switch (typid) {
	case BOOLOID:		/* boolean */
		return DatumGetBool(datum) ? Qtrue : Qfalse;
	case INT8OID:		/* bigint, bigserial */
		return LL2NUM(DatumGetInt64(datum));
	case INT2OID:		/* smallint, smallserial */
		return INT2NUM(DatumGetInt16(datum));
	case INT4OID:		/* integer, serial */
		return INT2NUM(DatumGetInt32(datum));

	case NUMERICOID:	/* decimal, numeric */
		klass = rb_const_get(rb_cObject, rb_intern("BigDecimal"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("BigDecimal is not CLASS");
		}
		dec = numeric_normalize(DatumGetNumeric(datum));
		arg[0] = rb_str_new_cstr(dec);
		return rb_class_new_instance(1, arg, klass);

	case FLOAT4OID:		/* real */
		return DBL2NUM(DatumGetFloat4(datum));
				/* Infinity、-Infinity、NaN を考慮すること */
	case FLOAT8OID:		/* double precision */
		return DBL2NUM(DatumGetFloat8(datum));

	case TEXTOID:		/* text */
		text_val = DatumGetTextP(datum);
		return string_force_encoding(rb_str_new(VARDATA(text_val),
					VARSIZE(text_val) - VARHDRSZ));

	case BPCHAROID:		/* character(n), char(n) */
		bp_char = DatumGetBpCharP(datum);
		return string_force_encoding(rb_str_new(VARDATA(bp_char),
					VARSIZE(bp_char) - VARHDRSZ));

	case VARCHAROID:	/* character varying(n), varchar(n) */
		var_char = DatumGetVarCharP(datum);
		return string_force_encoding(rb_str_new(VARDATA(var_char),
					VARSIZE(var_char) - VARHDRSZ));

	case BYTEAOID:		/* bytea */
		bytea_str = DatumGetByteaP(datum);
		return string_force_ascii_8bit(rb_str_new(VARDATA(bytea_str),
					VARSIZE(bytea_str) - VARHDRSZ));


	case DATEOID:		/* date */
		klass = rb_const_get(rb_cObject, rb_intern("Date"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("Date is not CLASS");
		}
		arg[0] = INT2NUM(DatumGetDateADT(datum) + 2451545);
				/*
				** datumの値は、2000-01-01を基準とした日数。
				** 2451545: 2000-01-01 のユリウス日
				*/
		t = rb_funcallv(klass, rb_intern("jd"), 1, arg);
		return t;

	case TIMEOID:		/* time [(p)] [without time zone] */
		arg[0] = INT2NUM((int) (DatumGetTimeADT(datum) / 1000000));
		arg[1] = INT2NUM((int) (DatumGetTimeADT(datum) % 1000000));
		t = rb_funcallv(rb_cTime, rb_intern("at"), 2, arg);
		t = rb_funcallv(t, rb_intern("utc"), 0, arg);
				/*
				** t = Time.at(sec, usec);
				** t.utc;
				*/
		return t;

	case TIMESTAMPOID:	/* timestamp [(p)] [without time zone] */
				/* Infinity、-Infinity を考慮すること */
		timestamp = DatumGetTimestamp(datum);
		timestamp2tm(timestamp, NULL, &tm, &fsec, NULL, NULL);
				/*
				** tzp (第2引数) がNULLなので、時間帯の
				** 調整はおこなわない。
				** tzn (第5引数) もNULLなので、時間帯名を
				** 返さない。
				*/
		klass = rb_const_get(rb_cObject, rb_intern("DateTime"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("DateTime is not CLASS");
		}
		arg[0] = INT2NUM(tm.tm_year);
		arg[1] = INT2NUM(tm.tm_mon);
		arg[2] = INT2NUM(tm.tm_mday);
		arg[3] = INT2NUM(tm.tm_hour);
		arg[4] = INT2NUM(tm.tm_min);
		arg[5] = INT2NUM(tm.tm_sec);
		t = rb_funcallv(klass, rb_intern("new"), 6, arg);
		return t;

	case TIMESTAMPTZOID:	/* timestamp [(p)] with time zone */
				/* Infinity、-Infinity を考慮すること */
		timestamptz = DatumGetTimestampTz(datum);
		timestamp2tm(timestamptz, &tzp, &tm, &fsec, NULL, NULL);
				/*
				** tzpを渡しているので時間帯の調整をおこなう。
				*/
		klass = rb_const_get(rb_cObject, rb_intern("DateTime"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("DateTime is not CLASS");
		}
		arg[0] = INT2NUM(tm.tm_year);
		arg[1] = INT2NUM(tm.tm_mon);
		arg[2] = INT2NUM(tm.tm_mday);
		arg[3] = INT2NUM(tm.tm_hour);
		arg[4] = INT2NUM(tm.tm_min);
		arg[5] = INT2NUM(tm.tm_sec);
		arg[6] = DBL2NUM((double) -tzp / 86400);  /* 時差: 日単位 */
		t = rb_funcallv(klass, rb_intern("new"), 7, arg);
		return t;

	case BITOID:		/* bit(n) */
	case VARBITOID:		/* bit varying(n) */
			/* FALLTHROUGH */
			/*
			** 「文字列」表現とし、次のようには実装しない。
			**	VarBit *	var_bit;
			**	var_bit = DatumGetVarBitP(datum);
			**	return rb_str_new(
			**		(const char *) VARBITS(var_bit),
			**		VARBITBYTES(var_bit));
			** (1) 長さが8の倍数でないとき、その旨を伝えられない。
			** (2) Ruby側のプログラムで扱いにくい。
			*/

	case INTERVALOID:	/* interval [fields] [(p)] */
			/* FALLTHROUGH */
			/*
			** 「文字列」表現とし、次のようには実装しない。
			**	Interval * span = DatumGetIntervalP(datum);
			**	interval2tm(*span, &tm, &fsec);
			**	t = Time.new tm.tm_{year,mon,mday,hour,min,sec}
			** (1) 負の時間間隔もありうるのでTimeでは表現できない。
			** (2) 「1か月」と「30日」を区別して扱う必要がある。
			*/
	default:
		getTypeOutputInfo(typid, &typOutput, &typIsVarlena);
		return string_force_encoding(
			rb_str_new_cstr(OidOutputFunctionCall(typOutput, datum)));
				/*
				** その他の場合は「文字列」表現。
				**
				** money
				** "char"	内部使用目的のため
				** name		内部使用目的のため
				** abstime	内部使用目的のため
				** reltime	内部使用目的のため
				** time with time zone	非推奨のため
				** enum
				** (幾何データ型)
				**	point, line, lseg, box, path,
				**	polygon, circle
				** (ネットワーク・アドレス型)
				**	cidr, inet, macaddr
				** (テキスト検索)
				**	tsvector、tsquery
				** uuid
				** xml
				** json, jsonb
				** (配列)
				** (範囲型)
				** (オブジェクト識別子)
				**	oid, regproc, regprocedure, regoper,
				**	regoperator, regclass, regtype,
				**	regconfig, regdictionary
				** pg_lsn
				** (擬似データ型)
				*/
	}
}

/* ---------------------------------------------------------------------
** 第n引数を取得し、型に応じて適切なrubyのオブジェクトに変換する。
*/
static
VALUE	NthArgToVALUE(FunctionCallInfo fcinfo, int n)
{
	Oid		typid;
	HeapTupleHeader	heap_tuple_header;
	Oid		tupType;
	int32		tupTypmod;
	TupleDesc	tupdesc;
	int		j;
	char *		attname;
	Oid		atttypeid;
	bool		is_null;

	VALUE		hash;
	Datum		datum;

	if (PG_ARGISNULL(n)) {
		return Qnil;
	}

	typid = get_fn_expr_argtype(fcinfo->flinfo, n);
	if (type_is_rowtype(typid)) {		/* 複合型の引数 */
		heap_tuple_header = PG_GETARG_HEAPTUPLEHEADER(n);

		tupType = HeapTupleHeaderGetTypeId(heap_tuple_header);
		tupTypmod = HeapTupleHeaderGetTypMod(heap_tuple_header);
		tupdesc = lookup_rowtype_tupdesc(tupType, tupTypmod);

		hash = rb_hash_new();
		for (j = 0; j < tupdesc->natts; j++) {
			if (tupdesc->attrs[j]->attisdropped) {
				continue;
			}
			attname = NameStr(tupdesc->attrs[j]->attname);
			atttypeid = tupdesc->attrs[j]->atttypid;
			datum = GetAttributeByName(
					heap_tuple_header, attname, &is_null);
			if (is_null) {
				rb_hash_aset(hash,
					rb_str_new_cstr(attname), Qnil);
			}
			else {
				rb_hash_aset(hash,
					rb_str_new_cstr(attname),
					DatumToVALUE(datum, atttypeid));
			}
		}

		ReleaseTupleDesc(tupdesc);

		return hash;
	}
	else {
		return DatumToVALUE(PG_GETARG_DATUM(n), typid);
	}
}

/* ---------------------------------------------------------------------
** VALUE (Ruby) → 文字列 → Datum (PostgreSQL) の変換 (文字列表現を経由)。
*/
static
Datum	VALUEtoDatumViaString(VALUE val, Oid typid)
{
	VALUE		t;
	char *		str;
	Oid		typInput;
	Oid		typIOParam;

	if (rb_respond_to(val, rb_intern("to_s"))) {
		t = rb_funcall(val, rb_intern("to_s"), 0);
	}
	else {
		t = rb_str_new_cstr("");
	}
	str = StringValuePtr(t);

	getTypeInputInfo(typid, &typInput, &typIOParam);
	return OidInputFunctionCall(typInput, str, typIOParam, -1);
}

/* ---------------------------------------------------------------------
** VALUE (Ruby) → Datum (PostgreSQL) の変換。
*/
static
Datum	VALUEtoDatum(VALUE val, Oid typid)
{
	VALUE		klass;
	VALUE		t;
	VALUE		arg[10];
	TimeADT		time_adt;
	struct pg_tm	tm;
	Timestamp	timestamp;
	TimestampTz	timestamptz;
	char *		str;
	int		len;
	bytea *		bytea_str;

	switch (typid) {
	case BOOLOID:		/* boolean */
		if ((val == Qnil) || (val == Qfalse)) {
			return BoolGetDatum(0);
		}
		else {
			return BoolGetDatum(1);
		}

	case INT8OID:		/* bigint, bigserial */
		if (rb_respond_to(val, rb_intern("to_i"))) {
			t = rb_funcallv(val, rb_intern("to_i"), 0, arg);
			return Int64GetDatum(NUM2LL(t));
		}
		else {
			return Int64GetDatum(0);
		}

	case INT2OID:		/* smallint, smallserial */
		if (rb_respond_to(val, rb_intern("to_i"))) {
			t = rb_funcallv(val, rb_intern("to_i"), 0, arg);
			return Int16GetDatum(NUM2INT(t));
		}
		else {
			return Int16GetDatum(0);
		}

	case INT4OID:		/* integer, serial */
		if (rb_respond_to(val, rb_intern("to_i"))) {
			t = rb_funcallv(val, rb_intern("to_i"), 0, arg);
			return Int32GetDatum(NUM2INT(t));
		}
		else {
			return Int32GetDatum(0);
		}

	case NUMERICOID:	/* decimal, numeric */
		klass = rb_const_get(rb_cObject, rb_intern("BigDecimal"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("BigDecimal is not CLASS");
		}
		/* val.is_kind_of(BigDecimal) の場合 */
		if (rb_obj_is_kind_of(val, klass) == Qtrue) {
			arg[0] = rb_str_new_cstr("F");
			t = rb_funcallv(val, rb_intern("to_s"), 1, arg);
			return VALUEtoDatumViaString(t, typid);
		}

		if (rb_respond_to(val, rb_intern("to_s"))) {
			t = rb_funcallv(val, rb_intern("to_s"), 0, arg);
			return VALUEtoDatumViaString(t, typid);
		}
		return VALUEtoDatumViaString(val, typid);
				/* 上記以外ならば「文字列」表現にして返す。 */

	case FLOAT4OID:		/* real */
				/* Infinity、-Infinity、NaN を考慮すること */
		if (rb_respond_to(val, rb_intern("to_f"))) {
			t = rb_funcallv(val, rb_intern("to_f"), 0, arg);
			return Float4GetDatum(NUM2DBL(t));
		}
		else {
			return Float4GetDatum(0);
		}

	case FLOAT8OID:		/* double precision */
				/* Infinity、-Infinity、NaN を考慮すること */
		if (rb_respond_to(val, rb_intern("to_f"))) {
			t = rb_funcallv(val, rb_intern("to_f"), 0, arg);
			return Float8GetDatum(NUM2DBL(t));
		}
		else {
			return Float8GetDatum(0);
		}

	case DATEOID:		/* date */
		/* val.is_kind_of(Date) の場合 */
		klass = rb_const_get(rb_cObject, rb_intern("Date"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("Date is not CLASS");
		}
		if (rb_obj_is_kind_of(val, klass) == Qtrue) {
			t = rb_funcallv(val, rb_intern("jd"), 0, arg);
			return DateADTGetDatum((DateADT) (NUM2INT(t) - 2451545));
				/*
				** datumの値は、2000-01-01を基準とした日数。
				** 2451545: 2000-01-01 のユリウス日
				*/
		}

		/* val.is_kind_of(Time) の場合 */
		if (rb_obj_is_kind_of(val, rb_cTime) == Qtrue) {
			arg[0] = INT2NUM(2000);
			t = rb_class_new_instance(1, arg, rb_cTime);
				/*
				** t = Time.new(2000);
				**  月以下はデフォルト値を使うので、
				**  2000-01-01 00:00:00 +09:00 になる。
				*/
			arg[0] = t;
			t = rb_funcallv(val, rb_intern("-"), 1, arg);
			return DateADTGetDatum((DateADT) (NUM2DBL(t) / (24 * 3600)));
		}
		return VALUEtoDatumViaString(val, typid);
				/* 上記以外ならば「文字列」表現にして返す。 */

	case TIMEOID:		/* time [(p)] [without time zone] */
		/* DateTimeなどのように、可能ならばTimeに変換しておく */
		if (rb_respond_to(val, rb_intern("to_time"))) {
			t = rb_funcall(val, rb_intern("to_time"), 0);
		}
		else {
			t = val;
		}

		/* t.is_kind_of(Time) の場合 */
		if (rb_obj_is_kind_of(t, rb_cTime) == Qtrue) {
			t = rb_funcall(t, rb_intern("utc"), 0);
				/*
				** 年月日の値は無視。
				*/
			time_adt = NUM2INT(rb_funcall(t, rb_intern("hour"), 0));
			time_adt *= 60;
			time_adt += NUM2INT(rb_funcall(t, rb_intern("min"), 0));
			time_adt *= 60;
			time_adt += NUM2INT(rb_funcall(t, rb_intern("sec"), 0));
			time_adt *= 1000000;
			time_adt += NUM2INT(rb_funcall(t, rb_intern("usec"), 0));
			return TimeADTGetDatum(time_adt);
		}
		return VALUEtoDatumViaString(val, typid);
				/* 上記以外ならば「文字列」表現にして返す。 */

	case TIMESTAMPOID:	/* timestamp [(p)] [without time zone] */
		/* val.is_kind_of(DateTime) || val.is_kind_of(Time) の場合 */
		klass = rb_const_get(rb_cObject, rb_intern("DateTime"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("DateTime is not CLASS");
		}
		if ((rb_obj_is_kind_of(val, klass) == Qtrue) ||
		    (rb_obj_is_kind_of(val, rb_cTime) == Qtrue)) {
			tm.tm_year = NUM2INT(rb_funcall(val, rb_intern("year"), 0));
			tm.tm_mon = NUM2INT(rb_funcall(val, rb_intern("mon"), 0));
			tm.tm_mday = NUM2INT(rb_funcall(val, rb_intern("day"), 0));
			tm.tm_hour = NUM2INT(rb_funcall(val, rb_intern("hour"), 0));
			tm.tm_min = NUM2INT(rb_funcall(val, rb_intern("min"), 0));
			tm.tm_sec = NUM2INT(rb_funcall(val, rb_intern("sec"), 0));
			tm2timestamp(&tm, 0, NULL, &timestamp);
			return TimestampGetDatum(timestamp);
		}

		return VALUEtoDatumViaString(val, typid);
				/* 上記以外ならば「文字列」表現にして返す。 */

	case TIMESTAMPTZOID:	/* timestamp [(p)] with time zone */
		/* val.is_kind_of(DateTime) || val.is_kind_of(Time) の場合 */
		klass = rb_const_get(rb_cObject, rb_intern("DateTime"));
		if (!RB_TYPE_P(klass, T_CLASS)) {
			info_out("DateTime is not CLASS");
		}
		if ((rb_obj_is_kind_of(val, klass) == Qtrue) ||
		    (rb_obj_is_kind_of(val, rb_cTime) == Qtrue)) {
			tm.tm_year = NUM2INT(rb_funcall(val, rb_intern("year"), 0));
			tm.tm_mon = NUM2INT(rb_funcall(val, rb_intern("mon"), 0));
			tm.tm_mday = NUM2INT(rb_funcall(val, rb_intern("day"), 0));
			tm.tm_hour = NUM2INT(rb_funcall(val, rb_intern("hour"), 0));
			tm.tm_min = NUM2INT(rb_funcall(val, rb_intern("min"), 0));
			tm.tm_sec = NUM2INT(rb_funcall(val, rb_intern("sec"), 0));
			if (rb_obj_is_kind_of(val, rb_cTime) == Qtrue) {
				tm.tm_gmtoff = NUM2INT(rb_funcall(val, rb_intern("gmtoff"), 0));
			}
			else {
				t = rb_funcall(val, rb_intern("offset"), 0);
				t = rb_funcall(t, rb_intern("*"), 1, INT2NUM(86400));
				t = rb_funcall(t, rb_intern("to_i"), 0);
				tm.tm_gmtoff = NUM2INT(t);
			}
/*
			info_out("[A] %04d-%02d-%02d %02d:%02d:%02d gmtoff = %ld",
				tm.tm_year, tm.tm_mon, tm.tm_mday,
				tm.tm_hour, tm.tm_min, tm.tm_sec,
				tm.tm_gmtoff);
*/
			tm2timestamp(&tm, 0, NULL, &timestamptz);
			return TimestampTzGetDatum(timestamptz);
		}

		return VALUEtoDatumViaString(val, typid);
				/* 上記以外ならば「文字列」表現にして返す。 */

	case BYTEAOID:		/* bytea */
		len = NUM2INT(rb_funcall(val, rb_intern("bytesize"), 0));
		str = StringValuePtr(val);

		bytea_str = (bytea *) palloc0(VARHDRSZ + len);
		SET_VARSIZE(bytea_str, VARHDRSZ + len);
		memcpy(VARDATA(bytea_str), str, len);
		return PointerGetDatum(bytea_str);


	case BITOID:		/* bit(n) */
	case VARBITOID:		/* bit varying(n) */
			/*
			** 「文字列」表現とするが、「0」以外の文字は
			** すべて「1」と看做す。
			*/
		if (rb_respond_to(val, rb_intern("tr"))) {
			arg[0] = rb_str_new_cstr("^0");
			arg[1] = rb_str_new_cstr("1");
			val = rb_funcallv(val, rb_intern("tr"), 2, arg);
		}
			/* FALLTHROUGH */

	case TEXTOID:		/* text */
	case BPCHAROID:		/* character(n), char(n) */
	case VARCHAROID:	/* character varying(n), varchar(n) */
	case INTERVALOID:	/* interval [fields] [(p)] */
			/* FALLTHROUGH */
			/*
			** 「文字列」表現とする。
			** PostgreSQLで認められている各種の書式が可能。
			*/
	default:
		return VALUEtoDatumViaString(val, typid);
				/*
				** その他の場合は「文字列」表現に変換して返す。
				*/
	}
}

/* ---------------------------------------------------------------------
** 戻り値がハッシュのとき、(PostgreSQL側の) 複合型戻り値として使う
** HeapTupleポインターの形に変換する。
*/
static
HeapTuple	HashToHeapTuple(VALUE val, TupleDesc tupdesc)
{
	Datum *		values;
	bool *		is_null;
	int		i;
	char *		attname;
	Oid		atttypid;
	VALUE		att_val;
	HeapTuple	heap_tuple;

	values = (Datum *) palloc(sizeof(Datum) * tupdesc->natts);
	is_null = (bool *) palloc(sizeof(bool) * tupdesc->natts);
	for (i = 0; i < tupdesc->natts; i++) {
		values[i] = (Datum) 0;
		is_null[i] = true;
		if (tupdesc->attrs[i]->attisdropped) {
			continue;
		}
		attname = NameStr(tupdesc->attrs[i]->attname);
		atttypid = tupdesc->attrs[i]->atttypid;
		att_val = rb_hash_aref(val, rb_str_new_cstr(attname));
		if (att_val != Qnil) {
			values[i] = VALUEtoDatum(att_val, atttypid);
			is_null[i] = false;
		}
	}
	heap_tuple = heap_form_tuple(tupdesc, values, is_null);
	pfree(is_null);
	pfree(values);

	return heap_tuple;
}

/* ---------------------------------------------------------------------
** 戻り値型に応じて、VALUE (Ruby) → Datum (PostgreSQL) の変換。
**	nil (PostgreSQL側のNULL) を返す特別扱い。
** 	(Ruby側のHashを) 複合型の戻り値に仕立てる処理。
*/
static
Datum	VALUEtoRetDatum(VALUE val, FunctionCallInfo fcinfo)
{
	Oid		rettype;
	TupleDesc	tupdesc;

	if (NIL_P(val)) {
		PG_RETURN_NULL();
	}

	rettype = get_fn_expr_rettype(fcinfo->flinfo);
	if (type_is_rowtype(rettype)) {		/* 複合型 */
		if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE) {
			info_out("func_class != TYPEFUNC_COMPOSITE");
			PG_RETURN_NULL();
		}
		if (rb_obj_is_kind_of(val, rb_cHash) == Qtrue) {
			return HeapTupleGetDatum(HashToHeapTuple(val, tupdesc));
		}
		else {
			PG_RETURN_NULL();
		}
	}
	else {
		return VALUEtoDatum(val, rettype);
	}
}

/* ---------------------------------------------------------------------
** 函数名、引数名、函数本体を取得し、rubyの函数定義文字列に仕立てる。
** func_name: (OUT) 函数名。
** is_named_args: (OUT) 引数名の指定の有無。この情報が不要ならばNULLを渡す。
** 戻り値: 函数定義文字列。次のような形。
**	# encoding: script_encoding
**	def func_name(arg1, arg2)
**		... scripts
**	end
**	引数名の指定がなければ、def func_name(opt = {}) となる。
**	script_encodingはpostgresql.confに、
**		plrb.script_encoding = '...'
**	の形で指定できる (既定値はutf-8)。
*/
static
char *	build_function_source(HeapTuple	procTup,
		char ** func_name, bool * is_named_args)
{
	const char *	script_encoding;
	bool		isnull;

	char *		def_func_source;	/* 函数定義文字列 */
	int		def_func_size = 100;	/* その長さ */
				/*
				** 初期値は「def」キーワードその他に割り当てる
				** 余裕分。これに函数本体、函数名、引数名の
				** 長さを加算して、必要な長さを求める。
				*/

	Datum		prosrc_datum;
	char *		proc_source;

	Datum		argnames_datum;
	Datum *		elems;
	int		nelems;
	int		j;

	/* -------------------------------------------------------------
	** スクリプト・エンコーディング。
	*/
	if ((script_encoding = GetConfigOption("plrb.script_encoding", true, false)) == NULL) {
		script_encoding = "utf-8";
	}

	/* -------------------------------------------------------------
	** 函数本体。
	*/
	prosrc_datum = SysCacheGetAttr(PROCOID, procTup,
			Anum_pg_proc_prosrc, &isnull);
	if (isnull) {
		elog(ERROR, "proc_prosrc is null");
	}
	proc_source = TextDatumGetCString(prosrc_datum);
	def_func_size += strlen(proc_source);

	/* -------------------------------------------------------------
	** 函数名。
	*/
	*func_name = NameStr(((Form_pg_proc) GETSTRUCT(procTup))->proname);
	def_func_size += strlen(*func_name);

	/* -------------------------------------------------------------
	** 引数名の並び。
	**	// get_func_arg_info() の実装に倣っている。
	**		// backend/utils/fmgr/funcapi.c 参照。
	*/
	argnames_datum = SysCacheGetAttr(PROCOID, procTup,
				Anum_pg_proc_proargnames, &isnull);
	if (! isnull) {		/* 引数名の指定がある場合 */
		deconstruct_array(DatumGetArrayTypeP(argnames_datum),
				TEXTOID, -1, false, 'i',
				&elems, NULL, &nelems);
		for (j = 0; j < nelems; j++) {
			def_func_size += (strlen(TextDatumGetCString(elems[j])) + 2);
					/* 2は ", " に必要な分 */
		}

		def_func_source = palloc(def_func_size);
		sprintf(def_func_source, "# encoding: %s\ndef %s(",
				script_encoding, *func_name);
		for (j = 0; j < nelems; j++) {
			strcat(def_func_source, TextDatumGetCString(elems[j]));
			if (j < nelems - 1) {
				strcat(def_func_source, ", ");
			}
		}
		strcat(def_func_source, ")\n");
		strcat(def_func_source, proc_source);
		if (def_func_source[strlen(def_func_source) - 1] != '\n') {
			strcat(def_func_source, "\n");
		}
		strcat(def_func_source, "end\n");
		if (is_named_args != NULL) {
			*is_named_args = true;
		}
	}
	else {		/* 引数名の指定がない場合 (トリガー函数を含む) */
		def_func_source = palloc(def_func_size);
		sprintf(def_func_source, "# encoding: %s\ndef %s(opt = {})\n",
				script_encoding, *func_name);
		strcat(def_func_source, proc_source);
		strcat(def_func_source, "\nend\n");
		if (is_named_args != NULL) {
			*is_named_args = false;
		}
	}

	return def_func_source;
}

/* =====================================================================
** ruby函数のコンパイル、引数の取得、実行。
** def_func_source: 「def func_name」～「end」までの函数定義文字列。
** func_name: 函数名。def_func_source中に現れている名前と同じであること。
** is_named_args: 引数名の指定があるか。
**	ない場合はoptという名前の配列で渡すものとして扱う。
** additional_arg: PostgreSQL側から渡された引数に追加する形で渡す引数。
**	トリガー函数のとき、トリガー情報を渡すために使う。
*/
static
VALUE	compile_and_execute(PG_FUNCTION_ARGS,
		char * def_func_source, char * func_name, bool is_named_args,
		VALUE additional_arg)
{
	VALUE		args;
	VALUE		val;
	int		n;
	int		rc;

	/* -------------------------------------------------------------
	** Ruby函数としてコンパイルする。
	** ScriptErrorが発生した場合はその旨のメッセージを出力する。
	*/
	rb_rescue2(x_rb_eval_string, rb_str_new_cstr(def_func_source),
			x_rb_eval_string_rescue, rb_str_new_cstr(func_name),
			rb_eScriptError, NULL);

	/* -------------------------------------------------------------
	** 引数を取得し、必要に応じて型変換を施し、Arrayに収容する。
	** この引数を渡して函数を呼び出す。
	*/
	args = rb_ary_new();
	rb_ary_push(args, rb_str_new_cstr(func_name));	/* 函数名 */
	for (n = 0; n < PG_NARGS(); n++) {
		rb_ary_push(args, NthArgToVALUE(fcinfo, n));
	}
	if (additional_arg != Qnil) {
		rb_ary_push(args, additional_arg);
	}

	rc = SPI_connect();
	if (rc != SPI_OK_CONNECT) {
		ereport(ERROR,
			(errcode(ERRCODE_CONNECTION_FAILURE),
			 errmsg("SPI_connect; Couldn't connect to SPI manager")));
	}
	if (is_named_args || additional_arg != Qnil) {
		val = rb_rescue(x_rb_funcallv, args, x_rb_rescue, rb_str_new_cstr(func_name));
	}
	else {
		val = rb_rescue(x_rb_funcall, args, x_rb_rescue, rb_str_new_cstr(func_name));
	};
	rc = SPI_finish();
	if (rc != SPI_OK_FINISH) {
		ereport(ERROR,
			(errcode(ERRCODE_CONNECTION_FAILURE),
			 errmsg("SPI_finish; Couldn't finish the SPI manager")));
	}

	return val;
}

/* ---------------------------------------------------------------------
** Pg::elog(level, report_string)
** エラーの報告
*/
static
VALUE	pg_elog(VALUE self, VALUE level, VALUE report_string)
{
	x_elog(NUM2INT(level), "%s", StringValuePtr(report_string));
	return Qnil;
}

/* ---------------------------------------------------------------------
** Pg::info_out(report_string)
**   Pg::elog(Pg::INFO, report_string) と同じ (デバッグ用の便宜)
*/
static
VALUE	pg_info_out(VALUE self, VALUE report_string)
{
	info_out("%s", StringValuePtr(report_string));
	return Qnil;
}

/* ---------------------------------------------------------------------
*/
#define	CanAccessResultRow(rc)	\
	(((rc) == SPI_OK_SELECT) || ((rc) == SPI_OK_INSERT_RETURNING) || ((rc) == SPI_OK_DELETE_RETURNING) || ((rc) == SPI_OK_UPDATE_RETURNING))

/* =====================================================================
** Pg::execute()
** SPIによるSQLコマンド実行はサブトランザクション内でおこなう。
** (参考: pltcl.c)
*/
/* ---------------------------------------------------------------------
*/
static
void	pg_subtrans_begin(MemoryContext oldcontext, ResourceOwner oldowner)
{
	BeginInternalSubTransaction(NULL);
	MemoryContextSwitchTo(oldcontext);
}

/* ---------------------------------------------------------------------
*/
static
void	pg_subtrans_commit(MemoryContext oldcontext, ResourceOwner oldowner)
{
	ReleaseCurrentSubTransaction();
	MemoryContextSwitchTo(oldcontext);
	CurrentResourceOwner = oldowner;

	SPI_restore_connection();
}

/* ---------------------------------------------------------------------
*/
static
void	pg_subtrans_abort(MemoryContext oldcontext, ResourceOwner oldowner)
{
	ErrorData *	edata;

	/* Save error info */
	MemoryContextSwitchTo(oldcontext);
	edata = CopyErrorData();
	FlushErrorState();

	/* Abort the inner transaction */
	RollbackAndReleaseCurrentSubTransaction();
	MemoryContextSwitchTo(oldcontext);
	CurrentResourceOwner = oldowner;

	SPI_restore_connection();

	FreeErrorData(edata);
}

/* ---------------------------------------------------------------------
** (大域変数である) SPI_tuptable、SPI_processed から、
** 第rec_numレコード (SPI函数の処理結果) を取り出し、Hashの形にする。
*/
static
VALUE	pg_exec_result_record(int rec_num)
{
	TupleDesc	tupdesc = SPI_tuptable->tupdesc;
	HeapTuple	row = SPI_tuptable->vals[rec_num];
				/*
				** SPI_tuptable: SPIの大域ポインター。
				*/
	VALUE		record;		/* Hash */
	int		j;
	char *		fname;
	Datum		datum;
	bool		is_null;

	record = rb_hash_new();
	for (j = 1; ; j++) {	/* 列番号は1から始まる */
		if ((fname = SPI_fname(tupdesc, j)) == NULL) {
			break;
		}
		datum = SPI_getbinval(row, tupdesc, j, &is_null);
		if (is_null) {
			rb_hash_aset(record,
				rb_str_new_cstr(fname), Qnil);
		}
		else {
			rb_hash_aset(record,
				rb_str_new_cstr(fname),
				DatumToVALUE(datum,
					SPI_gettypeid(tupdesc, j)));
		}
		pfree(fname);
	}
	return record;
}

/* ---------------------------------------------------------------------
** (大域変数である) SPI_tuptable、SPI_processed から、
** SPI函数の処理結果を取り出し、Array of Hashの形にする。
*/
static
VALUE	pg_exec_result()
{
	VALUE		result;		/* Array */
	int		rec_num;

	result = rb_ary_new();
	for (rec_num = 0; rec_num < (int) SPI_processed; rec_num++) {
			/*
			** SPI_processed: SPIの大域変数 (uint32)。
			** SQLコマンドが実行した実際の行数。
			*/
		rb_ary_push(result, pg_exec_result_record(rec_num));
	}
	return result;
}

/* ---------------------------------------------------------------------
** Pg::execute() の引数 (ruby側) を、PostgreSQL側の形式に変換する。
*/
static
int	exec_params_set(int argc, VALUE * argv, Oid **argtypes, Datum ** values, char ** nulls)
{
	VALUE		val;
	int		i;

	char *		str;
	int		len;
	text *		text_ptr;

	VALUE		k_bigdecimal = rb_const_get(rb_cObject, rb_intern("BigDecimal"));

	*argtypes = (Oid *) palloc0(sizeof(Oid) * argc);
	*values = (Datum *) palloc0(sizeof(Datum) * argc);
	*nulls = (char *) palloc0(sizeof(char) * argc);

	for (i = 0; i < argc; i++) {
		(*nulls)[i] = ' ';
		val = argv[i];
		if (val == Qnil) {
			(*nulls)[i] = 'n';
			(*argtypes)[i] = TEXTOID;
		}
		else if (val == Qtrue) {
			(*argtypes)[i] = BOOLOID;
			(*values)[i] = BoolGetDatum(1);
		}
		else if (val == Qfalse) {
			(*argtypes)[i] = BOOLOID;
			(*values)[i] = BoolGetDatum(0);
		}
		else if (rb_obj_is_kind_of(val, rb_cInteger) == Qtrue) {
			(*argtypes)[i] = INT8OID;
			(*values)[i] = Int64GetDatum(NUM2LL(val));
		}
		else if (rb_obj_is_kind_of(val, rb_cFloat) == Qtrue) {
			(*argtypes)[i] = FLOAT8OID;
			(*values)[i] = Float8GetDatum(NUM2DBL(val));
		}
		else {
			if (rb_obj_is_kind_of(val, k_bigdecimal) == Qtrue) {
				val = rb_funcall(val, rb_intern("to_s"), 1,
						rb_str_new_cstr("F"));
			}
			else if (rb_respond_to(val, rb_intern("to_s"))) {
				val = rb_funcall(val, rb_intern("to_s"), 0);
			}
			else {
				val = rb_str_new_cstr("");
			}

			(*argtypes)[i] = TEXTOID;
			str = StringValuePtr(val);
			len = NUM2INT(rb_funcall(val, rb_intern("bytesize"), 0));
				/* '\0' がありうるので len = strlen(str) は不可 */
			text_ptr = (text *) palloc0(VARHDRSZ + len);
			SET_VARSIZE(text_ptr, VARHDRSZ + len);
			memcpy(VARDATA(text_ptr), str, len);

			(*values)[i] = CStringGetDatum(text_ptr);
		}
	}
	return argc;
}

/* ---------------------------------------------------------------------
** Pg::execute(sql, arg1, arg2, ...);
** 戻り値: SELECTなど値を返すSQL文を実行した場合:
**	result = [
**		{ 'fname1' => 'value1', ... },
**		{ 'fname1' => 'value1', ... },
**	];
** 戻り値: 上記以外の場合: Qnil;
*/
static
VALUE	pg_execute(int argc, VALUE * argv, VALUE self)
{
	char *		sql;
	int		rc = 0;
	Oid *		argtypes;
	Datum *		values;
	char *		nulls;

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	if (argc == 0) {
		return Qnil;
	}

	sql = StringValuePtr(argv[0]);
	exec_params_set(argc - 1, &argv[1], &argtypes, &values, &nulls);

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		rc = SPI_execute_with_args(sql,
				argc - 1, argtypes, values, nulls,
				false, 0);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	pfree(nulls);
	pfree(values);
	pfree(argtypes);

	if (! CanAccessResultRow(rc) || (SPI_tuptable == NULL)) {
		return Qnil;
	}

	return pg_exec_result();
}

/* ---------------------------------------------------------------------
** [実験]
** Pg::funcall(function_name, arg1, arg2, ...);
** (PostgreSQL側で定義した) 函数を呼び出し、その結果を返す。
** 単一の値を返す函数に限る。
** // 実質的には次の処理と同じ。
**	Pg::select("SELECT #{func_name}($1, ...) AS a;", arg1, ...)[0]['a'];
*/
static
VALUE	pg_funcall(int argc, VALUE * argv, VALUE self)
{
	char *		func_name;
	char *		sql;
	int		i;
	char		buf[16];
	VALUE		val;

	if (argc == 0) {
		return Qnil;
	}

	func_name = StringValuePtr(argv[0]);
	sql = palloc(32 + strlen(func_name) + 5 * argc);

	sprintf(sql, "SELECT %s(", func_name);
	for (i = 1; i < argc; i++) {
		sprintf(buf, "$%d", i);
		strcat(sql, buf);
		if (i < argc - 1) {
			strcat(sql, ", ");
		}
	}
	strcat(sql, ") AS a;");
	argv[0] = rb_str_new_cstr(sql);

	val = pg_execute(argc, argv, self);
	pfree(sql);

	return rb_hash_aref(rb_ary_entry(val, 0), rb_str_new_cstr("a"));
}

/* ---------------------------------------------------------------------
** Pg::execute_with_type(sql, type1, arg1, type2, arg2, ...);
** Pg::execute()と違い、プレースホルダーの型を明示する。
** 戻り値: Pg::execute() と同じ。
*/
static
VALUE	pg_execute_with_type(int argc, VALUE * argv, VALUE self)
{
	char *		sql;
	int		rc = 0;
	int		num_args;
	Oid *		argtypes;
	Datum *		values;
	char *		nulls;
	int		i;
	VALUE		val;

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	if (argc == 0) {
		return Qnil;
	}

	sql = StringValuePtr(argv[0]);
	num_args = (argc - 1) / 2;
	argtypes = (Oid *) palloc0(sizeof(Oid) * num_args);
	values = (Datum *) palloc0(sizeof(Datum) * num_args);
	nulls = (char *) palloc0(sizeof(char) * num_args);

	for (i = 0; i < num_args; i++) {
		argtypes[i] = (Oid) NUM2INT(argv[i * 2 + 1]);
		val = argv[i * 2 + 2];
		if (val == Qnil) {
			nulls[i] = 'n';
		}
		else {
			nulls[i] = ' ';
			values[i] = VALUEtoDatum(val, argtypes[i]);
		}
	}

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		rc = SPI_execute_with_args(sql,
				num_args, argtypes, values, nulls,
				false, 0);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	pfree(nulls);
	pfree(values);
	pfree(argtypes);

	if (! CanAccessResultRow(rc) || (SPI_tuptable == NULL)) {
		return Qnil;
	}

	return pg_exec_result();
}

/* ---------------------------------------------------------------------
** Pg::prepare(sql, type1, type2, ...);
** 戻り値: プランナー
*/
static
VALUE	pg_prepare(int argc, VALUE * argv, VALUE self)
{
	char *		sql;
	Oid *		argtypes;
	int		i;
	int		cursorOptions = CURSOR_OPT_SCROLL;
	SPIPlanPtr	planPtr = NULL;

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	if (argc == 0) {
		return Qnil;
	}

	sql = StringValuePtr(argv[0]);
	argtypes = (Oid *) palloc0(sizeof(Oid) * (argc - 1));
	for (i = 0 ; i < argc - 1; i++) {
		argtypes[i] = (Oid) NUM2INT(argv[i + 1]);
	}

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		planPtr = SPI_prepare_cursor(sql,
				argc - 1, argtypes, cursorOptions);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	pfree(argtypes);

	return INT2NUM((int) planPtr);
}

/* ---------------------------------------------------------------------
** Pg::exec_prepared(plan, arg1, arg2, ...);
** plan: Pg::prepare() の戻り値。
** 戻り値: SELECTなど値を返すSQL文を実行した場合:
**	result = [
**		{ 'fname1' => 'value1', ... },
**		{ 'fname1' => 'value1', ... },
**	];
** 戻り値: 上記以外の場合: Qnil;
*/
static
VALUE	pg_exec_prepared(int argc, VALUE * argv, VALUE self)
{
	Datum *		values;
	char *		nulls;
	int		i;
	VALUE		val;
	SPIPlanPtr	plan;
	int		num_args;
	int		rc = 0;

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	if (argc == 0) {
		return Qnil;
	}

	plan = (SPIPlanPtr) NUM2INT(argv[0]);
	num_args = SPI_getargcount(plan);
	values = (Datum *) palloc0(sizeof(Datum) * num_args);
	nulls = (char *) palloc0(sizeof(char) * num_args);
	for (i = 0 ; i < num_args; i++) {
		val = (i + 1 < argc) ? argv[i + 1] : Qnil;
		if (val == Qnil) {
			nulls[i] = 'n';
		}
		else {
			nulls[i] = ' ';
			values[i] = VALUEtoDatum(val,
					SPI_getargtypeid(plan, i));
		}
	}

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		rc = SPI_execute_plan(plan, values, nulls,
				/* read_only = */ false,
				/* count = */ 0);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	pfree(nulls);
	pfree(values);

	if (! CanAccessResultRow(rc) || (SPI_tuptable == NULL)) {
		return Qnil;
	}

	return pg_exec_result();
}

/* ---------------------------------------------------------------------
** Pg::free_plan(plan);
** plan: Pg::prepare() の戻り値。
** 戻り値: Qnil;
*/
static
VALUE	pg_free_plan(VALUE self, VALUE plan_val)
{
	SPIPlanPtr	plan = (SPIPlanPtr) NUM2INT(plan_val);

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		SPI_freeplan(plan);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	return Qnil;
}

/* ---------------------------------------------------------------------
** Pg::open_cursor(plan, arg1, arg2, ...);
** 更新を伴うsql文は禁止する仕様 (read_only == true)。
** 戻り値: Portal
*/
static
VALUE	pg_open_cursor(int argc, VALUE * argv, VALUE self)
{
	SPIPlanPtr	plan;
	int		num_args;
	Datum *		values;
	char *		nulls;
	int		i;
	VALUE		val;
	Portal		portal = NULL;

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	if (argc == 0) {
		return Qnil;
	}

	plan = (SPIPlanPtr) NUM2INT(argv[0]);
	num_args = SPI_getargcount(plan);
	values = (Datum *) palloc0(sizeof(Datum) * num_args);
	nulls = (char *) palloc0(sizeof(char) * num_args);
	for (i = 0 ; i < num_args; i++) {
		val = (i + 1 < argc) ? argv[i + 1] : Qnil;
		if (val == Qnil) {
			nulls[i] = 'n';
		}
		else {
			nulls[i] = ' ';
			values[i] = VALUEtoDatum(val,
					SPI_getargtypeid(plan, i));
		}
	}

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		portal = SPI_cursor_open(NULL, plan,
				values, nulls,
				/* read_only = */ true);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	pfree(nulls);
	pfree(values);

	return INT2NUM((int) portal);
}

/* ---------------------------------------------------------------------
** Pg::fetch_one(portal);
** 戻り値: result = { 'fname1' => 'value1', ... } | nil
*/
static
VALUE	pg_fetch_one(VALUE self, VALUE portal_val)
{
	Portal		portal = (Portal) NUM2INT(portal_val);

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		SPI_scroll_cursor_fetch(portal, FETCH_FORWARD, 1);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	if ((SPI_processed == 1) && (SPI_tuptable != NULL)) {
		return pg_exec_result_record(0);
	}
	else {
		return Qnil;
	}
}

/* ---------------------------------------------------------------------
** Pg::close_cursol(portal);
** 戻り値: nil
*/
static
VALUE	pg_close_cursor(VALUE self, VALUE portal_val)
{
	Portal		portal = (Portal) NUM2INT(portal_val);

	MemoryContext	oldcontext = CurrentMemoryContext;
	ResourceOwner	oldowner = CurrentResourceOwner;

	pg_subtrans_begin(oldcontext, oldowner);
	PG_TRY();
	{
		SPI_cursor_close(portal);
		pg_subtrans_commit(oldcontext, oldowner);
	}
	PG_CATCH();
	{
		pg_subtrans_abort(oldcontext, oldowner);
	}
	PG_END_TRY();

	return Qnil;
}

/* ---------------------------------------------------------------------
** Pg::call_counter()
*/
static
VALUE	pg_call_counter(VALUE self)
{
	return rb_cv_get(self, "@@srf_call_counter");
}

/* ---------------------------------------------------------------------
** Pg::user_context()
*/
static
VALUE	pg_get_user_context(VALUE self)
{
	return rb_cv_get(self, "@@srf_user_context");
}

/* ---------------------------------------------------------------------
** Pg::user_context=()
*/
static
VALUE	pg_set_user_context(VALUE self, VALUE val)
{
	rb_cv_set(self, "@@srf_user_context", val);
	return self;
}

/* ---------------------------------------------------------------------
*/
static	VALUE	mod_pg = Qnil;

/* ---------------------------------------------------------------------
** (ruby側) Pgモジュールの初期化。
** SPI実行函数の定義、その他。
*/
static
void	pg_init(void)
{
	typedef	struct {
		const char *	name;
		Oid		typid;
	} TYPE_TBL;
	TYPE_TBL	type_tbl[] = {
		{ "TypeBOOL", BOOLOID },
			{ "TypeBOOLEAN", BOOLOID },
		{ "TypeINT8", INT8OID },
			{ "TypeBIGINT", INT8OID },
		{ "TypeINT2", INT2OID },
			{ "TypeSMALLINT", INT2OID },
		{ "TypeINT4", INT4OID },
			{ "TypeINTEGER", INT4OID },
		{ "TypeNUMERIC", NUMERICOID },
			{ "TypeDECIMAL", NUMERICOID },
		{ "TypeFLOAT4", FLOAT4OID },
			{ "TypeREAL", FLOAT4OID },
		{ "TypeFLOAT8", FLOAT8OID },
			{ "TypeDOUBLE", FLOAT8OID },
			{ "TypeDOUBLE_PRECISION", FLOAT8OID },
		{ "TypeTEXT", TEXTOID },
		{ "TypeBPCHAR", BPCHAROID },
			{ "TypeCHARACTER", BPCHAROID },
		{ "TypeVARCHAR", VARCHAROID },
			{ "TypeCHARACTER_VARYING", VARCHAROID },
		{ "TypeBYTEA", BYTEAOID },
		{ "TypeDATE", DATEOID },
		{ "TypeTIME", TIMEOID },
		{ "TypeTIMESTAMP", TIMESTAMPOID },
		{ "TypeTIMESTAMPTZ", TIMESTAMPTZOID },
		{ "TypeBIT", BITOID },
		{ "TypeVARBIT", VARBITOID },
			{ "TypeBIT_VARYING", VARBITOID },
		{ "TypeINTERVAL", INTERVALOID },
	};

	size_t		i;
	mod_pg = rb_define_module("Pg");

	rb_define_const(mod_pg, "PG_MAJOR", INT2NUM(1));
	rb_define_const(mod_pg, "PG_MINOR", INT2NUM(0));

	for (i = 0; i < sizeof(type_tbl) / sizeof(TYPE_TBL); i++) {
		rb_define_const(mod_pg, type_tbl[i].name, INT2NUM(type_tbl[i].typid));
	}

	rb_define_const(mod_pg, "FetchForward", INT2NUM(FETCH_FORWARD));
	rb_define_const(mod_pg, "FetchBackward", INT2NUM(FETCH_BACKWARD));
	rb_define_const(mod_pg, "FetchAbsolute", INT2NUM(FETCH_ABSOLUTE));
	rb_define_const(mod_pg, "FetchRelative", INT2NUM(FETCH_RELATIVE));

	rb_define_module_function(mod_pg, "execute", pg_execute, -1);
	rb_define_module_function(mod_pg, "select", pg_execute, -1);
	rb_define_module_function(mod_pg, "funcall", pg_funcall, -1);
	rb_define_module_function(mod_pg, "execute_with_type", pg_execute_with_type, -1);
	rb_define_module_function(mod_pg, "select_with_type", pg_execute_with_type, -1);
	rb_define_module_function(mod_pg, "prepare", pg_prepare, -1);
	rb_define_module_function(mod_pg, "exec_prepared", pg_exec_prepared, -1);
	rb_define_module_function(mod_pg, "free_plan", pg_free_plan, 1);
	rb_define_module_function(mod_pg, "open_cursor", pg_open_cursor, -1);
	rb_define_module_function(mod_pg, "fetch_one", pg_fetch_one, 1);
	rb_define_module_function(mod_pg, "close_cursor", pg_close_cursor, 1);

	rb_define_const(mod_pg, "LevelDEBUG5", INT2NUM(DEBUG5));
	rb_define_const(mod_pg, "LevelDEBUG4", INT2NUM(DEBUG4));
	rb_define_const(mod_pg, "LevelDEBUG3", INT2NUM(DEBUG3));
	rb_define_const(mod_pg, "LevelDEBUG2", INT2NUM(DEBUG2));
	rb_define_const(mod_pg, "LevelDEBUG1", INT2NUM(DEBUG1));
	rb_define_const(mod_pg, "LevelLOG", INT2NUM(LOG));
	rb_define_const(mod_pg, "LevelCOMMERROR", INT2NUM(COMMERROR));
	rb_define_const(mod_pg, "LevelINFO", INT2NUM(INFO));
	rb_define_const(mod_pg, "LevelNOTICE", INT2NUM(NOTICE));
	rb_define_const(mod_pg, "LevelWARNING", INT2NUM(WARNING));
	rb_define_const(mod_pg, "LevelERROR", INT2NUM(ERROR));
	rb_define_const(mod_pg, "LevelFATAL", INT2NUM(FATAL));
	rb_define_const(mod_pg, "LevelPANIC", INT2NUM(PANIC));

	rb_define_module_function(mod_pg, "elog", pg_elog, 2);
	rb_define_module_function(mod_pg, "info_out", pg_info_out, 1);

	rb_cv_set(mod_pg, "@@srf_call_counter", INT2NUM(0));
	rb_define_module_function(mod_pg, "call_counter", pg_call_counter, 0);

	rb_cv_set(mod_pg, "@@srf_user_context", Qnil);
	rb_define_module_function(mod_pg, "user_context", pg_get_user_context, 0);
	rb_define_module_function(mod_pg, "user_context=", pg_set_user_context, 1);
}

/* ---------------------------------------------------------------------
** PLrb函数 (非トリガー) の入口点 (集合を返す函数の場合)。
*/
static
Datum	function_call_handler_setof(PG_FUNCTION_ARGS, HeapTuple procTup)
{
	char *		def_func_source = NULL;
	char *		func_name;
	bool		is_named_args;
	VALUE		val = Qnil;

	FuncCallContext *	funcctx;
	VALUE		val_n;
	Datum		result;

	def_func_source = build_function_source(
			procTup, &func_name, &is_named_args);

	if (SRF_IS_FIRSTCALL()) {
		MemoryContext	oldcontext;
		funcctx = SRF_FIRSTCALL_INIT();
		rb_cv_set(mod_pg, "@@srf_call_counter",
			INT2NUM(funcctx->call_cntr));
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		val = compile_and_execute(fcinfo,
			def_func_source, func_name, is_named_args, Qnil);

		if (rb_obj_is_kind_of(val, rb_cArray) == Qtrue) {
			/* Arrayモード */
			/* 集合が配列の形で一括して返された場合 */
			funcctx->user_fctx = (void *) val;
			funcctx->max_calls =
				NUM2INT(rb_funcall(val, rb_intern("size"), 0));
		}
		else {
			/* Fetchモード */
			/* 集合の第1要素のみ返された場合 */
			funcctx->user_fctx = NULL;
			funcctx->max_calls = 0;
		}

		MemoryContextSwitchTo(oldcontext);
	}
	funcctx = SRF_PERCALL_SETUP();

	if (funcctx->user_fctx != NULL) {	/* Arrayモード */
		pfree(def_func_source);
		if (funcctx->call_cntr < funcctx->max_calls) {
			val = (VALUE) funcctx->user_fctx;
			val_n = rb_funcall(val, rb_intern("[]"), 1,
					INT2NUM(funcctx->call_cntr));
			result = VALUEtoRetDatum(val_n, fcinfo);
			SRF_RETURN_NEXT(funcctx, result);
		}
		else {
			SRF_RETURN_DONE(funcctx);
		}
	}
	else {					/* Fetchモード */
		if (0 < funcctx->call_cntr) {
			MemoryContext	oldcontext;
			rb_cv_set(mod_pg, "@@srf_call_counter",
				INT2NUM(funcctx->call_cntr));
			oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
			val_n = compile_and_execute(fcinfo,
				def_func_source, func_name, is_named_args,
				Qnil);
			MemoryContextSwitchTo(oldcontext);
		}
		else {
			val_n = val;
		}
		pfree(def_func_source);
		if (val_n != Qnil) {
			result = VALUEtoRetDatum(val_n, fcinfo);
			SRF_RETURN_NEXT(funcctx, result);
		}
		else {
			SRF_RETURN_DONE(funcctx);
		}
	}
}

/* ---------------------------------------------------------------------
** PLrb函数 (非トリガー) の入口点 (集合を返す函数ではない場合)。
*/
static
Datum	function_call_handler_no_setof(PG_FUNCTION_ARGS, HeapTuple procTup)
{
	char *		def_func_source;
	char *		func_name;
	bool		is_named_args;
	VALUE		val;

	def_func_source = build_function_source(
			procTup, &func_name, &is_named_args);
	val = compile_and_execute(fcinfo,
			def_func_source, func_name, is_named_args, Qnil);
	pfree(def_func_source);

	if (NIL_P(val)) {
		PG_RETURN_NULL();
	}
	else {
		return VALUEtoRetDatum(val, fcinfo);
	}
}

/* =====================================================================
** PLrb函数 (非トリガー) の入口点。
*/
static
Datum	function_call_handler(PG_FUNCTION_ARGS)
				/* FunctionCallInfo fcinfo */
{
	HeapTuple	procTup;
	bool		is_retset;
	Datum		result;

	procTup = SearchSysCache1(PROCOID,
			ObjectIdGetDatum(fcinfo->flinfo->fn_oid));
	is_retset = ((Form_pg_proc) GETSTRUCT(procTup))->proretset;

	if (is_retset) {			/* 集合を返す函数 */
		result = function_call_handler_setof(fcinfo, procTup);
	}
	else {
		result = function_call_handler_no_setof(fcinfo, procTup);
	}
	ReleaseSysCache(procTup);
	return result;
}

/* ---------------------------------------------------------------------
** (ruby側) Tgモジュールの初期化の下請け。
**	Tg::RECORD_OLD、Tg::RECORD_NEW (レコード値のハッシュ) を生成する。
*/
static
VALUE	tg_init_tuple_data(TupleDesc tupdesc, HeapTuple heap_tuple)
{
	VALUE		rec;
	int		i;
	char *		attname;
	Oid		atttypid;
	Datum		attr_datum;
	bool		is_null;

	if (heap_tuple == NULL) {
		return Qnil;
	}

	rec = rb_hash_new();
	for (i = 0; i < tupdesc->natts; i++) {
		if (tupdesc->attrs[i]->attisdropped) {
			continue;
		}
		attname = NameStr(tupdesc->attrs[i]->attname);
		atttypid = tupdesc->attrs[i]->atttypid;
		attr_datum = heap_getattr(heap_tuple, i + 1, tupdesc, &is_null);
		if (is_null) {
			rb_hash_aset(rec, rb_str_new_cstr(attname), Qnil);
		}
		else {
			rb_hash_aset(rec, rb_str_new_cstr(attname),
					DatumToVALUE(attr_datum, atttypid));
		}
	}
	return rec;
}

/* ---------------------------------------------------------------------
** (ruby側) トリガー・データを収容するハッシュの生成
**	WHEN = BEFORE | AFTER | INSTEAD
**	LEVEL = ROW | STATEMENT
**	EVENT = INSERT | UPDATE | DELETE | TRUNCATE
**	RECORD_OLD = 削除しようとする行、または更新前の行 (Hash)
**	RECORD_NEW = 挿入しようとする行、または更新後の行 (Hash)
**	NAME = トリガー名
**	TABLE_NAME = トリガーの発行元テーブル名
**	TABLE_SCHEMA = トリガーの発行元テーブルがあるスキーマの名前
**	ARGC = トリガー函数の引数の数
**	ARGV = トリガー函数の引数の配列
*/
static
VALUE	trig_data_to_hash(TriggerData * trig_data)
{
	VALUE		hash = rb_hash_new();
	TupleDesc	tupdesc = trig_data->tg_relation->rd_att;
	int		i;
	VALUE		argv;

	/* -------------------------------------------------------------
	*/
	if (TRIGGER_FIRED_BEFORE(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("WHEN"), rb_str_new_cstr("BEFORE"));
	}
	else if (TRIGGER_FIRED_AFTER(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("WHEN"), rb_str_new_cstr("AFTER"));
	}
	else if (TRIGGER_FIRED_INSTEAD(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("WHEN"), rb_str_new_cstr("INSTEAD"));
	}
	else {
		rb_hash_aset(hash, rb_str_new_cstr("WHEN"), rb_str_new_cstr("UNKNOWN"));
	}

	/* -------------------------------------------------------------
	*/
	if (TRIGGER_FIRED_FOR_ROW(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("LEVEL"), rb_str_new_cstr("ROW"));
	}
	else if (TRIGGER_FIRED_FOR_STATEMENT(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("LEVEL"), rb_str_new_cstr("STATEMENT"));
	}
	else {
		rb_hash_aset(hash, rb_str_new_cstr("LEVEL"), rb_str_new_cstr("UNKNOWN"));
	}

	/* -------------------------------------------------------------
	*/
	if (TRIGGER_FIRED_BY_INSERT(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("EVENT"), rb_str_new_cstr("INSERT"));
		if (TRIGGER_FIRED_FOR_ROW(trig_data->tg_event)) {
			rb_hash_aset(hash, rb_str_new_cstr("RECORD_NEW"),
				tg_init_tuple_data(tupdesc, trig_data->tg_trigtuple));
		}
	}
	else if (TRIGGER_FIRED_BY_UPDATE(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("EVENT"), rb_str_new_cstr("UPDATE"));
		if (TRIGGER_FIRED_FOR_ROW(trig_data->tg_event)) {
			rb_hash_aset(hash, rb_str_new_cstr("RECORD_OLD"),
				tg_init_tuple_data(tupdesc, trig_data->tg_trigtuple));
			rb_hash_aset(hash, rb_str_new_cstr("RECORD_NEW"),
				tg_init_tuple_data(tupdesc, trig_data->tg_newtuple));
		}
	}
	else if (TRIGGER_FIRED_BY_DELETE(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("EVENT"), rb_str_new_cstr("DELETE"));
		if (TRIGGER_FIRED_FOR_ROW(trig_data->tg_event)) {
			rb_hash_aset(hash, rb_str_new_cstr("RECORD_OLD"),
				tg_init_tuple_data(tupdesc, trig_data->tg_trigtuple));
		}
	}
	else if (TRIGGER_FIRED_BY_TRUNCATE(trig_data->tg_event)) {
		rb_hash_aset(hash, rb_str_new_cstr("EVENT"), rb_str_new_cstr("TRUNCATE"));
	}
	else {
		rb_hash_aset(hash, rb_str_new_cstr("EVENT"), rb_str_new_cstr("UNKNOWN"));
	}

	/* -------------------------------------------------------------
	*/
	rb_hash_aset(hash, rb_str_new_cstr("TRIGGER_NAME"),
		rb_str_new_cstr(trig_data->tg_trigger->tgname));
	rb_hash_aset(hash, rb_str_new_cstr("TABLE_NAME"),
		rb_str_new_cstr(SPI_getrelname(trig_data->tg_relation)));
	rb_hash_aset(hash, rb_str_new_cstr("TABLE_SCHEMA"),
		rb_str_new_cstr(SPI_getnspname(trig_data->tg_relation)));

	/* -------------------------------------------------------------
	** トリガー函数の引数
	*/
	rb_hash_aset(hash, rb_str_new_cstr("ARGC"),
		INT2NUM(trig_data->tg_trigger->tgnargs));
	argv = rb_ary_new();
	for (i = 0; i < trig_data->tg_trigger->tgnargs; i++) {
		rb_ary_push(argv, rb_str_new_cstr(trig_data->tg_trigger->tgargs[i]));
	}
	rb_hash_aset(hash, rb_str_new_cstr("ARGV"), argv);

	return hash;
}

/* =====================================================================
** トリガー函数の入口点。
*/
static
Datum	trigger_handler(PG_FUNCTION_ARGS)
{
	HeapTuple	procTup;
	char *		def_func_source;
	char *		func_name;
	VALUE		val;

	/* -------------------------------------------------------------
	*/
	TriggerData *	trig_data = (TriggerData *) fcinfo->context;
	VALUE		trig_hash = trig_data_to_hash(trig_data);
	TupleDesc	tupdesc = trig_data->tg_relation->rd_att;

	/* -------------------------------------------------------------
	*/
	procTup = SearchSysCache1(PROCOID,
			ObjectIdGetDatum(fcinfo->flinfo->fn_oid));
	def_func_source = build_function_source(procTup, &func_name, NULL);
	ReleaseSysCache(procTup);
	val = compile_and_execute(fcinfo,
			def_func_source, func_name, false, trig_hash);
	pfree(def_func_source);

	if (rb_obj_is_kind_of(val, rb_cHash) == Qtrue) {
		return PointerGetDatum(HashToHeapTuple(val, tupdesc));
	}
	else {
		PG_RETURN_VOID();
	}
}

/* ---------------------------------------------------------------------
** (ruby側) イヴェント・トリガー・データを収容するハッシュの生成
**	EVENT = ddl_command_start | ddl_command_end | sql_drop
**	TAG = 実行対象イヴェントに関連するコマンド・タグ。
**		"CREATE FUNCTION" など。
*/
static
VALUE	event_trig_data_to_hash(EventTriggerData * event_trig_data)
{
	VALUE		hash = rb_hash_new();

	rb_hash_aset(hash, rb_str_new_cstr("EVENT"),
		rb_str_new_cstr(event_trig_data->event));
	rb_hash_aset(hash, rb_str_new_cstr("TAG"),
		rb_str_new_cstr(event_trig_data->tag));

	return hash;
}

/* =====================================================================
** イヴェント・トリガー函数の入口点。
*/
static
Datum	event_trigger_handler(PG_FUNCTION_ARGS)
{
	HeapTuple	procTup;
	char *		def_func_source;
	char *		func_name;
	VALUE		val;

	/* -------------------------------------------------------------
	*/
	EventTriggerData *	event_trig_data =
				(EventTriggerData *) fcinfo->context;
	VALUE		event_trig_hash =
				event_trig_data_to_hash(event_trig_data);

	/* -------------------------------------------------------------
	*/
	procTup = SearchSysCache1(PROCOID,
			ObjectIdGetDatum(fcinfo->flinfo->fn_oid));
	def_func_source = build_function_source(procTup, &func_name, NULL);
	ReleaseSysCache(procTup);
	val = compile_and_execute(fcinfo,
			def_func_source, func_name, false, event_trig_hash);
	pfree(def_func_source);

	PG_RETURN_VOID();
}


/* ---------------------------------------------------------------------
** postgresql.conf に
**	plrb.auto_require = 'foo,baa'
** のような記述があれば、自動的に (PostgreSQLの函数定義に記述しなくても)
** fooおよびbaaをrequire()する。
*/
static
void	auto_require()
{
	const char *	conf_val;
	char *		buf;
	char *		pt;

	if ((conf_val = GetConfigOption("plrb.auto_require", true, false)) == NULL) {
		return;
	}
	buf = malloc(strlen(conf_val) + 1);
	strcpy(buf, conf_val);
	for (pt = strtok(buf, ", "); pt != NULL; pt = strtok(NULL, ", ")) {
		rb_rescue2(x_rb_require, rb_str_new_cstr(pt),
			x_rb_require_rescue, rb_str_new_cstr(pt),
			rb_eLoadError, NULL);
	}
	free(buf);
}

/* =====================================================================
*/
static	bool			g_already_ruby_inited = false;
static	FunctionCallInfo	g_plrb_current_fcinfo = NULL;

/* ---------------------------------------------------------------------
** ruby処理系の初期化 / Pgモジュールの定義。
*/
static
void	plrb_ruby_initialize()
{
	if (! g_already_ruby_inited) {
		ruby_init();
		ruby_init_loadpath();
		pg_init();
		rb_require("bigdecimal");
		rb_require("date");
		auto_require();
		g_already_ruby_inited = true;
	}
}

/* =====================================================================
** PLrbの入口点。
*/
PG_FUNCTION_INFO_V1(plrb_call_handler);
Datum	plrb_call_handler(PG_FUNCTION_ARGS)
		/* FunctionCallInfo fcinfo */
{
	Datum			retval;
	FunctionCallInfo	save_fcinfo;

	/* -------------------------------------------------------------
	*/
	plrb_ruby_initialize();

	save_fcinfo = g_plrb_current_fcinfo;

	PG_TRY();
	{
		if (CALLED_AS_TRIGGER(fcinfo)) {
			g_plrb_current_fcinfo = NULL;
			retval = trigger_handler(fcinfo);
		}
		else if (CALLED_AS_EVENT_TRIGGER(fcinfo)) {
			g_plrb_current_fcinfo = NULL;
			event_trigger_handler(fcinfo);
			retval = (Datum) 0;
		}
		else {
			g_plrb_current_fcinfo = fcinfo;
			retval = function_call_handler(fcinfo);
		}
	}
	PG_CATCH();
	{
		g_plrb_current_fcinfo = save_fcinfo;
		PG_RE_THROW();
	}
	PG_END_TRY();

	g_plrb_current_fcinfo = save_fcinfo;

	return retval;
}

/* =====================================================================
** PLrb Validatorの入口点。
*/
PG_FUNCTION_INFO_V1(plrb_validator);
Datum	plrb_validator(PG_FUNCTION_ARGS)
		/* FunctionCallInfo fcinfo */
{
	Oid		funcoid = PG_GETARG_OID(0);
	HeapTuple	procTup;
	char *		def_func_source;
	char *		func_name;

	if (! CheckFunctionValidatorAccess(fcinfo->flinfo->fn_oid, funcoid)) {
		PG_RETURN_VOID();
	}

	if (! check_function_bodies) {		/* utils/guc.h に宣言がある */
		PG_RETURN_VOID();
	}

	/* -------------------------------------------------------------
	*/
	plrb_ruby_initialize();

	/* -------------------------------------------------------------
	*/
	procTup = SearchSysCache1(PROCOID, ObjectIdGetDatum(funcoid));
			/*
			** plrb_call_handler() の場合とは違い、引数として
			** 渡されたfuncoidを使ってdef_func_sourceを取得する。
			*/
	def_func_source = build_function_source(procTup, &func_name, NULL);
	ReleaseSysCache(procTup);

	/* -------------------------------------------------------------
	** Ruby函数としてコンパイルする。
	** ScriptErrorが発生した場合はその旨のメッセージを出力する。
	*/
	rb_rescue2(x_rb_eval_string, rb_str_new_cstr(def_func_source),
			x_rb_eval_string_rescue, rb_str_new_cstr(func_name),
			rb_eScriptError, NULL);

	pfree(def_func_source);

	PG_RETURN_VOID();
}

/* =====================================================================
** PLrb inline_handlerの入口点。
*/
PG_FUNCTION_INFO_V1(plrb_inline_handler);
Datum	plrb_inline_handler(PG_FUNCTION_ARGS)
		/* FunctionCallInfo fcinfo */
{
	InlineCodeBlock *	codeblock = (InlineCodeBlock *) PG_GETARG_POINTER(0);
	char *	anon_func_name = (char *) "anon_func";
	char *		def_func_source = codeblock->source_text;
	FunctionCallInfo	save_fcinfo;
	int		sav_nargs;

	/* -------------------------------------------------------------
	*/
	plrb_ruby_initialize();

	/* -------------------------------------------------------------
	*/
	def_func_source = palloc(100 + strlen(codeblock->source_text));
	sprintf(def_func_source, "def %s()\n%s\nend\n",
			anon_func_name, codeblock->source_text);
			/*
			** plrb_call_handler() の場合とは違い、
			** 引数codeblock経由でコード文字列を取得する。
			*/

	/* -------------------------------------------------------------
	*/
	save_fcinfo = g_plrb_current_fcinfo;
	PG_TRY();
	{
		g_plrb_current_fcinfo = fcinfo;
		sav_nargs = fcinfo->nargs;
		fcinfo->nargs = 0;
				/*
				** internal型引数をひとつ取るため1になって
				** いるが、0であるものとして実行する。
				*/
		compile_and_execute(fcinfo,
			def_func_source, anon_func_name, true, Qnil);
		fcinfo->nargs = sav_nargs;
	}
	PG_CATCH();
	{
		g_plrb_current_fcinfo = save_fcinfo;
		PG_RE_THROW();
	}
	PG_END_TRY();

	pfree(def_func_source);
	g_plrb_current_fcinfo = save_fcinfo;

	PG_RETURN_VOID();
}

