#! /usr/local/bin/ruby
# encoding: utf-8
#
# test_func_args.rb
# 函数のさまざまな引数型に応じ、適切なRubyのオブジェクトを生成する旨のテスト。
#

require_relative 'am_assert';
include AmAssert;
require_relative 'pgx';

# ======================================================================
#
class TestFuncArgs
	# ==============================================================
	#
	def initialize()
		@pq = Pq.new('postgresql://localhost/db_plrb');
	end

	# ==============================================================
	# 引数名がない形で定義された函数の実行。戻り値型はinteger。
	#
	def test_call_func_by_anonymous_args()
		test_cases = [
			[ '(a)', 5, 'succ', 6 ],
			[ '(b)', 5, 'double', 10 ],
			[ '(c)', 5, 'ELSE', 5 ],
		];
		sql = "SELECT func_by_anonymous_args($1, $2) AS val;";
		test_cases.each do |id, args_0, args_1, guess_val|
			result = @pq.select_one(sql, args_0, args_1);
			am_assert_equals(guess_val, result['val'], "func_by_anonymous_args: #{id}");
		end
	end

	# ==============================================================
	# 引数名つきで定義された函数の実行。戻り値型はtext。
	#
	def test_call_func_by_named_args()
		test_cases = [
			[ '(a)', 5, 'succ', '5 -> 6' ],
			[ '(b)', 5, 'double', '5 -> 10' ],
			[ '(c)', 5, 'ELSE', 'Unknown Operator ELSE' ],
			[ '(d)', 5, '反転', '5 -> -5' ],
			[ '(e)', 5, '演算子', 'Unknown Operator 演算子' ],
		];
		sql = "SELECT func_by_named_args($1, $2) AS val;";
		test_cases.each do |id, curr, op, guess_val|
			result = @pq.select_one(sql, curr, op);
			actual_val = result['val'].force_encoding("UTF-8");
				# pgx.rb はエンコーディングに関して何もしない。
				# そのままではASCII-8BITになっている。
				#
			am_assert_equals(guess_val, result['val'], "func_by_named_args: #{id}");
		end
	end

	# ==============================================================
	# NULLを渡したときRubyのnilになること
	#
	def test_null_args()
		sql = "SELECT null_args(NULL) AS dt;";
		result = @pq.select_one(sql);
		am_assert_equals("arg is nil", result['dt']);
	end

	# ==============================================================
	# 64ビット整数 (bigint) 型
	#
	def test_bigint_args()
		test_cases = [
			[ 4294967295, "big: 4294967296" ],
			[ 4294967296, "big: 4294967297" ],
			[ 386124294967295, "big: 386124294967296" ],
			[ 9223372036854775807, "big: 9223372036854775808" ],
						## max int64
			[ -9223372036854775808, "big: -9223372036854775807" ],
						## min int64
			[ -4294967296, "big: -4294967295" ],
		];
		sql = "SELECT bigint_args($1) AS dt;";
		test_cases.each do |a_num, guess_result|
			result = @pq.select_one(sql, a_num);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# NUMERIC (decimal) 型
	#
	def test_numeric_args()
		test_cases = [
			[ 9223372036854775807, "numeric: 9223372036854775808.0" ],
						## max int64
			[ -9223372036854775808, "numeric: -9223372036854775807.0" ],
						## min int64
		];
		sql = "SELECT numeric_args($1) AS dt;";
		test_cases.each do |a_num, guess_result|
			result = @pq.select_one(sql, a_num);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 浮動小数点数型
	#
	def test_float_args()
		test_cases = [
			# real, double precision, guess_result
			[ 3.4, 5.6, "a_num * 2:  6.8, b_num * 2: 11.2" ],
			[ 'Infinity', 'Infinity', "a_num * 2:  Inf, b_num * 2:  Inf" ],
			[ '-Infinity', '-Infinity', "a_num * 2: -Inf, b_num * 2: -Inf" ],
			[ 'NaN', 'NaN', "a_num * 2:  NaN, b_num * 2:  NaN" ],
		];
		sql = "SELECT float_args($1::real, $2::double precision) AS dt;";
		test_cases.each do |a_ch, b_ch, guess_result|
			result = @pq.select_one(sql, a_ch, b_ch);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 文字列型 (固定長/可変長)
	#
	def test_char_args()
		test_cases = [
			# char(5), varchar(8), guess_result
			[ "abCDE", "defghIJK", "a_ch: ABCDE, b_ch: DEFGHIJK" ],
			[ "abC", "deF", "a_ch: ABC  , b_ch: DEF" ],
			[ "abcDEFgH", "defgHIJKlm", "a_ch: ABCDE, b_ch: DEFGHIJK" ],
			[ "日", "日本", "a_ch: 日  , b_ch: 日本" ],
		];
		sql = "SELECT character_args($1::char(5), $2::varchar(8)) AS dt;";
		test_cases.each do |a_ch, b_ch, guess_result|
			result = @pq.select_one(sql, a_ch, b_ch);
			dt = result['dt'];
			dt.force_encoding('UTF-8');
			am_assert_equals(guess_result, dt);
		end
	end

	# ==============================================================
	# バイト列型
	#
	def test_bytea_args()
		test_cases = [
			[ '\\\\x20FF0043C0FFEE', 'a_str: 20:FF:00:43:C0:FF:EE:' ],
		];
		test_cases.each do |inp, guess_result|
			sql = "SELECT bytea_args(E'#{inp}'::bytea) AS dt;";
			result = @pq.select_one(sql);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 日付型
	#
	def test_date_args()
		test_cases = [
			[ "2015-10-30", "d: 2015-10-30" ],
			[ "2000-01-02", "d: 2000-01-02" ],
			[ "1982-01-02", "d: 1982-01-02" ],
			[ "1845-02-04", "d: 1845-02-04" ],
		];
		sql = "SELECT date_args($1::date) AS dt;";
		test_cases.each do |date_string, guess_result|
			result = @pq.select_one(sql, date_string);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時刻型
	#
	def test_time_args()
		test_cases = [
			[ "00:00:00", "t: 1970-01-01 00:00:00.000 +0000" ],
			[ "00:00:01.789", "t: 1970-01-01 00:00:01.789 +0000" ],
			[ "13:24:56", "t: 1970-01-01 13:24:56.000 +0000" ],
		];
		sql = "SELECT time_args($1::time) AS dt;";
		test_cases.each do |time_string, guess_result|
			result = @pq.select_one(sql, time_string);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型 (時間帯なし)
	#
	def test_timestamp_args()
		test_cases = [
			[ "2015-05-21 13:24:32", "tomorrow: Friday; 2015-05-22 13:24:32 +0000" ],
		];
		sql = "SELECT timestamp_args($1::timestamp) AS dt;";
		test_cases.each do |a_ch, guess_result|
			result = @pq.select_one(sql, a_ch);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型 (時間帯あり)
	#
	def test_timestamptz_args()
		test_cases = [
			[ "2015-05-21 13:24:32 JST", "tomorrow: Friday; 2015-05-22 13:24:32 +0900" ],	## JST: +0900
			[ "2015-05-21 13:24:32 PST", "tomorrow: Saturday; 2015-05-23 06:24:32 +0900" ],	## PST: -0800
			[ "2015-05-21 13:24:32 Asia/Jakarta", "tomorrow: Friday; 2015-05-22 15:24:32 +0900" ],	## Asia/Jakarta: +0700
		];
		sql = "SELECT timestamptz_args($1::timestamptz) AS dt;";
		test_cases.each do |a_ch, guess_result|
			result = @pq.select_one(sql, a_ch);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# ビット列型 (固定長/可変長)
	#
	def test_bit_args()
		test_cases = [
			# bit(12), bit varying(12), guess_result
			[ "010000110101", "010000110101", %Q|BIT: a_str: "010000110101", b_str: "010000110101"| ],
			# B'0100 0011' = 0x43 = 'C'
			# B'0101 0000' = 0x50 = 'P' // 後半の '0000' は埋め草
			[ "010000110101", "01000011", %Q|BIT: a_str: "010000110101", b_str: "01000011"| ],
		];
		sql = "SELECT bit_args($1::bit(12), $2::bit varying(12)) AS dt;";
		test_cases.each do |a_ch, b_ch, guess_result|
			result = @pq.select_one(sql, a_ch, b_ch);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時間間隔型
	#
	def test_interval_args()
		test_cases = [
			[ "1 year 2 months 3 days 4 hours 5 minutes 6 seconds", "span: 1 year 2 mons 3 days 04:05:06" ],
			[ "P1Y2M3DT4H5M6S", "span: 1 year 2 mons 3 days 04:05:06" ],
			[ "3 months ago", "span: -3 mons" ], ## 負の時間間隔。
		];
		sql = "SELECT interval_args($1::interval) AS dt;";
		test_cases.each do |a_ch, guess_result|
			result = @pq.select_one(sql, a_ch);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 複合型引数 (1) テーブル名を複合型の名前と看做して使用
	#
	def test_music_tbl_arg()
		@pq.exec("INSERT INTO  music_tbl (name, weight) VALUES ('Oboe', 500);");

		test_cases = [
			[ '(1)', 'Oboe', 300, 'OBOE: 600' ],
		];
		sql = "SELECT name, music_tbl_arg(music_tbl, $1) AS r FROM music_tbl WHERE name = $2;";
		test_cases.each do |id, name, val, guess_result|
			result = @pq.select_one(sql, val, name);
			am_assert_equals(guess_result, result['r']);
		end
	end

	# ==============================================================
	# 複合型引数 (2)
	#
	def test_trio_arg()
		test_cases = [
			[ 10, 5, 'Sonata', 't1 * 3 + t2: 35; t3: SONATA' ],
		];
		sql = "SELECT trio_arg(ROW($1, $2, $3)) AS r;";
		test_cases.each do |t1, t2, t3, guess_result|
			result = @pq.select_one(sql, t1, t2, t3);
			am_assert_equals(guess_result, result['r']);
		end
	end

end

# ======================================================================
#
if __FILE__ == $0
	t = TestFuncArgs.new();
	t.test_call_func_by_anonymous_args();
	t.test_call_func_by_named_args();

	t.test_null_args();

	# bool
	t.test_bigint_args();
	# smallint
	# integer
	t.test_numeric_args();
	t.test_float_args();
	# text
	t.test_char_args();
	t.test_bytea_args();
	t.test_date_args();
	t.test_time_args();
	t.test_timestamp_args();
	t.test_timestamptz_args();
	t.test_bit_args();
	t.test_interval_args();
	t.test_music_tbl_arg();
	t.test_trio_arg();
end

