#! /usr/local/bin/ruby
# encoding: utf-8
#
# test_func_types.rb
# 函数の戻り値型に応じ、Rubyのオブジェクトを適切に変換して返す旨のテスト。
#

require_relative 'am_assert';
include AmAssert;
require_relative 'pgx';

# ======================================================================
#
class TestFuncTypes
	# ==============================================================
	#
	def initialize()
		@pq = Pq.new('postgresql://localhost/db_plrb');
	end

	# ==============================================================
	# text型だが常にnil。これをNULLとして扱うこと。
	#
	def test_func_always_nil()
		sql = <<-"__SQL__";
			SELECT	CASE WHEN func_always_nil('foo') IS NULL THEN
					'is null'
				ELSE
					'NOT null'
				END AS judge ;
		__SQL__
		result = @pq.select_one(sql);
		am_assert_equals('is null', result['judge']);
	end

	# ==============================================================
	# 整数型: しかし函数の実体は文字列を返す形 (to_iで自動変換する)。
	#
	def test_func_int_returns_str()
		test_cases = [
			[ "2345", 2345 ],
			[ "-34SA", -34 ],
			[ "NOT NUMBER", 0 ],
		];
		sql = "SELECT func_int_returns_str($1) AS ival;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['ival']);
		end
	end

	# ==============================================================
	# 整数型: しかし函数の実体はSymbol (to_iがない) を返す形。
	#
	def test_func_int_returns_symbol()
		test_cases = [
			[ "2345", 0 ],
			[ "DUMMY", 0 ],
		];
		sql = "SELECT func_int_returns_symbol($1) AS ival;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['ival']);
		end
	end

	# ==============================================================
	# NUMERIC (decimal) 型: 函数の実体はIntegerを返す形。
	#
	def test_func_decimal_returns_int()
		test_cases = [
			[ "9223372036854775807", BigDecimal.new("9223372036854775807") ],
			[ "-9223372036854775808", BigDecimal.new("-9223372036854775808") ],
			[ "0.9223372036854775807", BigDecimal.new("0") ],
					# Integerなので0になってしまう
			[ "1119223372036854775807", BigDecimal.new("1119223372036854775807") ],
					# Integerの範囲を超える値
		];
		sql = "SELECT func_decimal_returns_int($1) AS dt;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# NUMERIC (decimal) 型: 函数の実体はFloatを返す形。
	#
	def test_func_decimal_returns_float()
		test_cases = [
			[ "9223372036854775807", BigDecimal.new("9223372036854776000") ],
			[ "-9223372036854775808", BigDecimal.new("-9223372036854776000") ],
			[ "0.9223372036854775807", BigDecimal.new("0.9223372036854776") ],
			[ "1119223372036854775807", BigDecimal.new("1119223372036854800000") ],
					# 精度が落ちている
		];
		sql = "SELECT func_decimal_returns_float($1) AS dt;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# NUMERIC (decimal) 型: 函数の実体はBigDecimalを返す形。
	#
	def test_func_decimal_returns_bigdecimal()
		test_cases = [
			[ "9223372036854775807", BigDecimal.new("9223372036854775807.0") ],
			[ "-9223372036854775808", BigDecimal.new("-9223372036854775808.0") ],
			[ "0.9223372036854775807", BigDecimal.new("0.9223372036854775807") ],
			[ "1119223372036854775807", BigDecimal.new("1119223372036854775807.0") ],
					# Integerの範囲を超える値
		];
		sql = "SELECT func_decimal_returns_bigdecimal($1) AS dt;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# バイト列型: 函数の実体はStringを返す形。
	def test_func_bytea_returns_str()
		test_cases = [
			[ "C1FABE00FD1F", "\\xc1fabe00fd1f" ],
		];
		sql = "SELECT func_bytea_returns_str($1) AS dt;";
		test_cases.each do |str, guess_result|
			result = @pq.select_one(sql, str);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 日付型: 函数の実体はDateを返す形。
	#
	def test_func_date_returns_date()
		test_cases = [
			[ 2000, 1, 2, Date.parse("2000-01-03") ],
			[ 2015, 5, 31, Date.parse("2015-06-01") ],
		];
		sql = "SELECT func_date_returns_date($1, $2, $3) AS dt;";
		test_cases.each do |y, m, d, guess_result|
			result = @pq.select_one(sql, y, m, d);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 日付型: 函数の実体はTimeを返す形。
	#
	def test_func_date_returns_time()
		test_cases = [
			[ 2000, 1, 2, Date.parse("2000-01-03") ],
			[ 2015, 5, 31, Date.parse("2015-06-01") ],
		];
		sql = "SELECT func_date_returns_time($1, $2, $3) AS dt;";
		test_cases.each do |y, m, d, guess_result|
			result = @pq.select_one(sql, y, m, d);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 日付型: 函数の実体は文字列 ("YYYY-MM-DD") を返す形。
	#
	def test_func_date_returns_str()
		test_cases = [
			[ 2000, 1, 2, Date.parse("2000-01-03") ],
			[ 2015, 5, 30, Date.parse("2015-05-31") ],
		];
				# func_date_returns_str() の実装は手抜きで、
				# 月末日の翌日を正しく求められない。
		sql = "SELECT func_date_returns_str($1, $2, $3) AS dt;";
		test_cases.each do |y, m, d, guess_result|
			result = @pq.select_one(sql, y, m, d);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時刻型: 函数の実体はDateTimeを返す形。
	#
	def test_func_time_returns_datetime()
		test_cases = [
			[ 1, 0, 1, Time.utc(1970, 1, 1, 1, 0, 1) ],
		];
		sql = "SELECT func_time_returns_datetime($1, $2, $3) AS dt;";
		test_cases.each do |h, m, s, guess_result|
			result = @pq.select_one(sql, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時刻型: 函数の実体はTimeを返す形。
	#
	def test_func_time_returns_time()
		test_cases = [
			[ 2, 0, 1, Time.utc(1970, 1, 1, 2, 0, 1) ],
		];
		sql = "SELECT func_time_returns_time($1, $2, $3) AS dt;";
		test_cases.each do |h, m, s, guess_result|
			result = @pq.select_one(sql, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時刻型: 函数の実体は文字列 ("HH:MM:SS") を返す形。
	#
	def test_func_time_returns_str()
		test_cases = [
			[ 3, 0, 1, Time.utc(1970, 1, 1, 3, 0, 1) ],
		];
		sql = "SELECT func_time_returns_str($1, $2, $3) AS dt;";
		test_cases.each do |h, m, s, guess_result|
			result = @pq.select_one(sql, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型: 函数の実体はDateTimeを返す形。
	#
	def test_func_timestamp_returns_datetime()
		test_cases = [
			[ 2015, 1, 3, 1, 0, 1, DateTime.parse("2015-01-03 01:00:01") ],
			[ 1512, 1, 3, 1, 0, 1, DateTime.parse("1512-01-03 01:00:01") ],
		];
		sql = "SELECT func_timestamp_returns_datetime($1, $2, $3, $4, $5, $6) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型: 函数の実体はTimeを返す形。
	#
	def test_func_timestamp_returns_time()
		test_cases = [
			[ 2015, 1, 3, 1, 0, 1, DateTime.parse("2015-01-03 01:00:01") ],
			[ 1512, 1, 3, 1, 0, 1, DateTime.parse("1512-01-03 01:00:01") ],
		];
		sql = "SELECT func_timestamp_returns_time($1, $2, $3, $4, $5, $6) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型: 函数の実体は文字列を返す形。
	#
	def test_func_timestamp_returns_str()
		test_cases = [
			[ 2015, 1, 3, 1, 0, 1, DateTime.parse("2015-01-03 01:00:01") ],
			[ 1512, 1, 3, 1, 0, 1, DateTime.parse("1512-01-03 01:00:01") ],
		];
		sql = "SELECT func_timestamp_returns_str($1, $2, $3, $4, $5, $6) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型 (TZ): 函数の実体はDateTimeを返す形。
	#
	def test_func_timestamptz_returns_datetime()
		test_cases = [
			[ 2015, 1, 3, 1, 0, 1, 5 * 3600, DateTime.parse("2015-01-03 10:00:01+09") ],
		];
		sql = "SELECT func_timestamptz_returns_datetime($1, $2, $3, $4, $5, $6, $7) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, gmtoff, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s, gmtoff);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型 (TZ): 函数の実体はTimeを返す形。
	#
	def test_func_timestamptz_returns_time()
		test_cases = [
			[ 2015, 1, 5, 1, 0, 1, 5 * 3600, DateTime.parse("2015-01-05 10:00:01+09") ],
		];
		sql = "SELECT func_timestamptz_returns_time($1, $2, $3, $4, $5, $6, $7) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, gmtoff, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s, gmtoff);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# タイムスタンプ型 (TZ): 函数の実体は文字列を返す形。
	#
	def test_func_timestamptz_returns_str()
		test_cases = [
			[ 2015, 1, 7, 1, 0, 1, 5 * 3600, DateTime.parse("2015-01-07 05:00:01+09") ],
		];
		sql = "SELECT func_timestamptz_returns_str($1, $2, $3, $4, $5, $6, $7) AS dt;";
		test_cases.each do |year, mon, day, h, m, s, gmtoff, guess_result|
			result = @pq.select_one(sql, year, mon, day, h, m, s, gmtoff);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# ビット列型
	#
	def test_func_bit_returning_str()
		test_cases = [
			[ "101100010010", "101100010010" ],
			[ "AAAA0000AAAA", "111100001111" ],
					# '0' 以外はすべて '1' と看做す
		];
		sql = "SELECT func_bit_returns_str($1) AS dt;";
		test_cases.each do |inp, guess_result|
			result = @pq.select_one(sql, inp);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 時間間隔型
	#
	def test_func_interval()
		test_cases = [
			[ 15, 1, 2, "15 years 1 mon 2 days" ],
		];
		sql = "SELECT func_interval($1, $2, $3) AS dt;";
		test_cases.each do |y, m, d, guess_result|
			result = @pq.select_one(sql, y, m, d);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 複合型(1)
	# CREATE TYPE で定義した型
	#
	def test_func_trio()
		test_cases = [
			[ 7, %Q|(7,14,"a3 is 21")| ],
		];
		sql = "SELECT func_trio($1) AS dt;";
		test_cases.each do |mid, guess_result|
			result = @pq.select_one(sql, mid);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 複合型(2)
	# CREATE TABLE に伴い暗黙に定義された型
	#
	def test_func_complex()
		test_cases = [
			[ 'Bassoon', 230, '(,bassoon,f,94.5,460,)' ],
		];
		sql = "SELECT func_complex($1, $2) AS dt;";
		test_cases.each do |name, weight, guess_result|
			result = @pq.select_one(sql, name, weight);
			am_assert_equals(guess_result, result['dt']);
		end
	end

	# ==============================================================
	# 集合を返すこと(1)
	# Hashの配列で返す形
	#
	def test_func_setof_trio()
		test_cases = [
			[ 7, 
				[
					{ 'dt' => %Q|(7,14,"a3 is 21")| },
					{ 'dt' => %Q|(42,49,"a3 is 56")| },
				]
			],
		];
		sql = "SELECT func_setof_trio($1) AS dt;";
		test_cases.each do |mid, guess_result|
			result = @pq.select_all(sql, mid);
			am_assert_equals(guess_result, result);
		end
	end

	# ==============================================================
	# 集合を返すこと(2)
	# 呼び出しの都度、1レコードずつ返す形
	#
	def test_func_setof_trio_fetch()
		test_cases = [
			[ 7, 
				[
					{ 'dt' => %Q|(7,14,"a3 is 21")| },
					{ 'dt' => %Q|(42,49,"a3 is 56")| },
				]
			],
		];
		sql = "SELECT func_setof_trio_fetch($1) AS dt;";
		test_cases.each do |mid, guess_result|
			result = @pq.select_all(sql, mid);
			am_assert_equals(guess_result, result);
		end
	end

end

# ======================================================================
#
if __FILE__ == $0
	t = TestFuncTypes.new();

	t.test_func_always_nil();

	t.test_func_int_returns_str();
	t.test_func_int_returns_symbol();

	t.test_func_decimal_returns_int();
	t.test_func_decimal_returns_float();
	t.test_func_decimal_returns_bigdecimal();

	t.test_func_bytea_returns_str();

	t.test_func_date_returns_date();
	t.test_func_date_returns_time();
	t.test_func_date_returns_str();

	t.test_func_time_returns_datetime();
	t.test_func_time_returns_time();
	t.test_func_time_returns_str();

	t.test_func_timestamp_returns_datetime();
	t.test_func_timestamp_returns_time();
	t.test_func_timestamp_returns_str();

	t.test_func_timestamptz_returns_datetime();
	t.test_func_timestamptz_returns_time();
	t.test_func_timestamptz_returns_str();

	t.test_func_bit_returning_str();

	t.test_func_interval();

	t.test_func_trio();
	t.test_func_complex();
	t.test_func_setof_trio();
	t.test_func_setof_trio_fetch();
end

